#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2006-2010 (ita)

"ocaml support"

import os, re
from waflib import Utils, Task
from waflib.Logs import error
from waflib.TaskGen import feature, before_method, after_method, extension

EXT_MLL = ['.mll']
EXT_MLY = ['.mly']
EXT_MLI = ['.mli']
EXT_MLC = ['.c']
EXT_ML  = ['.ml']

open_re = re.compile(r'^\s*open\s+([a-zA-Z]+)(;;){0,1}$', re.M)
foo = re.compile(r"""(\(\*)|(\*\))|("(\\.|[^"\\])*"|'(\\.|[^'\\])*'|.[^()*"'\\]*)""", re.M)
def filter_comments(txt):
	meh = [0]
	def repl(m):
		if m.group(1):
			meh[0] += 1
		elif m.group(2):
			meh[0] -= 1
		elif not meh[0]:
			return m.group()
		return ''
	return foo.sub(repl, txt)

def scan(self):
	node = self.inputs[0]
	code = filter_comments(node.read())

	global open_re
	names = []
	import_iterator = open_re.finditer(code)
	if import_iterator:
		for import_match in import_iterator:
			names.append(import_match.group(1))
	found_lst = []
	raw_lst = []
	for name in names:
		nd = None
		for x in self.incpaths:
			nd = x.find_resource(name.lower()+'.ml')
			if not nd:
				nd = x.find_resource(name+'.ml')
			if nd:
				found_lst.append(nd)
				break
		else:
			raw_lst.append(name)

	return (found_lst, raw_lst)

native_lst=['native', 'all', 'c_object']
bytecode_lst=['bytecode', 'all']

@feature('ocaml')
def init_ml(self):
	Utils.def_attrs(self,
		type = 'all',
		incpaths_lst = [],
		bld_incpaths_lst = [],
		mlltasks = [],
		mlytasks = [],
		mlitasks = [],
		native_tasks = [],
		bytecode_tasks = [],
		linktasks = [],
		bytecode_env = None,
		native_env = None,
		compiled_tasks = [],
		includes = '',
		uselib = '',
		are_deps_set = 0)

@feature('ocaml')
@after_method('init_ml')
def init_envs_ml(self):

	self.islibrary = getattr(self, 'islibrary', False)

	global native_lst, bytecode_lst
	self.native_env = None
	if self.type in native_lst:
		self.native_env = self.env.derive()
		if self.islibrary:
			self.native_env['OCALINKFLAGS']   = '-a'

	self.bytecode_env = None
	if self.type in bytecode_lst:
		self.bytecode_env = self.env.derive()
		if self.islibrary:
			self.bytecode_env['OCALINKFLAGS'] = '-a'

	if self.type == 'c_object':
		self.native_env.append_unique('OCALINKFLAGS_OPT', '-output-obj')

@feature('ocaml')
@before_method('apply_vars_ml')
@after_method('init_envs_ml')
def apply_incpaths_ml(self):
	inc_lst = self.includes.split()
	lst = self.incpaths_lst
	for dir in inc_lst:
		node = self.path.find_dir(dir)
		if not node:
			error("node not found: " + str(dir))
			continue
		if not node in lst:
			lst.append(node)
		self.bld_incpaths_lst.append(node)
	# now the nodes are added to self.incpaths_lst

@feature('ocaml')
@before_method('process_source')
def apply_vars_ml(self):
	for i in self.incpaths_lst:
		if self.bytecode_env:
			app = self.bytecode_env.append_value
			app('OCAMLPATH', ['-I', i.bldpath(), '-I', i.srcpath()])

		if self.native_env:
			app = self.native_env.append_value
			app('OCAMLPATH', ['-I', i.bldpath(), '-I', i.srcpath()])

	varnames = ['INCLUDES', 'OCAMLFLAGS', 'OCALINKFLAGS', 'OCALINKFLAGS_OPT']
	for name in self.uselib.split():
		for vname in varnames:
			cnt = self.env[vname+'_'+name]
			if cnt:
				if self.bytecode_env:
					self.bytecode_env.append_value(vname, cnt)
				if self.native_env:
					self.native_env.append_value(vname, cnt)

@feature('ocaml')
@after_method('process_source')
def apply_link_ml(self):

	if self.bytecode_env:
		ext = self.islibrary and '.cma' or '.run'

		linktask = self.create_task('ocalink')
		linktask.bytecode = 1
		linktask.set_outputs(self.path.find_or_declare(self.target + ext))
		linktask.env = self.bytecode_env
		self.linktasks.append(linktask)

	if self.native_env:
		if self.type == 'c_object':
			ext = '.o'
		elif self.islibrary:
			ext = '.cmxa'
		else:
			ext = ''

		linktask = self.create_task('ocalinkx')
		linktask.set_outputs(self.path.find_or_declare(self.target + ext))
		linktask.env = self.native_env
		self.linktasks.append(linktask)

		# we produce a .o file to be used by gcc
		self.compiled_tasks.append(linktask)

@extension(*EXT_MLL)
def mll_hook(self, node):
	mll_task = self.create_task('ocamllex', node, node.change_ext('.ml'))
	mll_task.env = self.native_env.derive()
	self.mlltasks.append(mll_task)

	self.source.append(mll_task.outputs[0])

@extension(*EXT_MLY)
def mly_hook(self, node):
	mly_task = self.create_task('ocamlyacc', node, [node.change_ext('.ml'), node.change_ext('.mli')])
	mly_task.env = self.native_env.derive()
	self.mlytasks.append(mly_task)
	self.source.append(mly_task.outputs[0])

	task = self.create_task('ocamlcmi', mly_task.outputs[1], mly_task.outputs[1].change_ext('.cmi'))
	task.env = self.native_env.derive()

@extension(*EXT_MLI)
def mli_hook(self, node):
	task = self.create_task('ocamlcmi', node, node.change_ext('.cmi'))
	task.env = self.native_env.derive()
	self.mlitasks.append(task)

@extension(*EXT_MLC)
def mlc_hook(self, node):
	task = self.create_task('ocamlcc', node, node.change_ext('.o'))
	task.env = self.native_env.derive()
	self.compiled_tasks.append(task)

@extension(*EXT_ML)
def ml_hook(self, node):
	if self.native_env:
		task = self.create_task('ocamlx', node, node.change_ext('.cmx'))
		task.env = self.native_env.derive()
		task.incpaths = self.bld_incpaths_lst
		self.native_tasks.append(task)

	if self.bytecode_env:
		task = self.create_task('ocaml', node, node.change_ext('.cmo'))
		task.env = self.bytecode_env.derive()
		task.bytecode = 1
		task.incpaths = self.bld_incpaths_lst
		self.bytecode_tasks.append(task)

def compile_may_start(self):

	if not getattr(self, 'flag_deps', ''):
		self.flag_deps = 1

		# the evil part is that we can only compute the dependencies after the
		# source files can be read (this means actually producing the source files)
		if getattr(self, 'bytecode', ''):
			alltasks = self.generator.bytecode_tasks
		else:
			alltasks = self.generator.native_tasks

		self.signature() # ensure that files are scanned - unfortunately
		tree = self.generator.bld
		for node in self.inputs:
			lst = tree.node_deps[self.uid()]
			for depnode in lst:
				for t in alltasks:
					if t == self:
						continue
					if depnode in t.inputs:
						self.set_run_after(t)

		# TODO necessary to get the signature right - for now
		delattr(self, 'cache_sig')
		self.signature()

	return Task.Task.runnable_status(self)

class ocamlx(Task.Task):
	"""native caml compilation"""
	color   = 'GREEN'
	run_str = '${OCAMLOPT} ${OCAMLPATH} ${OCAMLFLAGS} ${OCAMLINCLUDES} -c -o ${TGT} ${SRC}'
	scan    = scan
	runnable_status = compile_may_start

class ocaml(Task.Task):
	"""bytecode caml compilation"""
	color   = 'GREEN'
	run_str = '${OCAMLC} ${OCAMLPATH} ${OCAMLFLAGS} ${OCAMLINCLUDES} -c -o ${TGT} ${SRC}'
	scan    = scan
	runnable_status = compile_may_start

class ocamlcmi(Task.Task):
	"""interface generator (the .i files?)"""
	color   = 'BLUE'
	run_str = '${OCAMLC} ${OCAMLPATH} ${OCAMLINCLUDES} -o ${TGT} -c ${SRC}'
	before  = ['ocamlcc', 'ocaml', 'ocamlcc']

class ocamlcc(Task.Task):
	"""ocaml to c interfaces"""
	color   = 'GREEN'
	run_str = 'cd ${TGT[0].bld_dir()} && ${OCAMLOPT} ${OCAMLFLAGS} ${OCAMLPATH} ${OCAMLINCLUDES} -c ${SRC[0].abspath()}'

class ocamllex(Task.Task):
	"""lexical generator"""
	color   = 'BLUE'
	run_str = '${OCAMLLEX} ${SRC} -o ${TGT}'
	before  = ['ocamlcmi', 'ocaml', 'ocamlcc']

class ocamlyacc(Task.Task):
	"""parser generator"""
	color   = 'BLUE'
	run_str = '${OCAMLYACC} -b ${tsk.base()} ${SRC}'
	before  = ['ocamlcmi', 'ocaml', 'ocamlcc']

	def base(self):
		node = self.outputs[0]
		s = os.path.splitext(node.name)[0]
		return node.bld_dir() + os.sep + s

def link_may_start(self):

	if getattr(self, 'bytecode', 0):
		alltasks = self.generator.bytecode_tasks
	else:
		alltasks = self.generator.native_tasks

	for x in alltasks:
		if not x.hasrun:
			return Task.ASK_LATER

	if not getattr(self, 'order', ''):

		# now reorder the inputs given the task dependencies
		# this part is difficult, we do not have a total order on the tasks
		# if the dependencies are wrong, this may not stop
		seen = []
		pendant = []+alltasks
		while pendant:
			task = pendant.pop(0)
			if task in seen:
				continue
			for x in task.run_after:
				if not x in seen:
					pendant.append(task)
					break
			else:
				seen.append(task)
		self.inputs = [x.outputs[0] for x in seen]
		self.order = 1
	return Task.Task.runnable_status(self)

class ocalink(Task.Task):
	"""bytecode caml link"""
	color   = 'YELLOW'
	run_str = '${OCAMLC} -o ${TGT} ${OCAMLINCLUDES} ${OCALINKFLAGS} ${SRC}'
	runnable_status = link_may_start
	after = ['ocaml', 'ocamlcc']

class ocalinkx(Task.Task):
	"""native caml link"""
	color   = 'YELLOW'
	run_str = '${OCAMLOPT} -o ${TGT} ${OCAMLINCLUDES} ${OCALINKFLAGS_OPT} ${SRC}'
	runnable_status = link_may_start
	after = ['ocamlx', 'ocamlcc']

def configure(conf):
	opt = conf.find_program('ocamlopt', var='OCAMLOPT', mandatory=False)
	occ = conf.find_program('ocamlc', var='OCAMLC', mandatory=False)
	if (not opt) or (not occ):
		conf.fatal('The objective caml compiler was not found:\ninstall it or make it available in your PATH')

	v = conf.env
	v['OCAMLC']       = occ
	v['OCAMLOPT']     = opt
	v['OCAMLLEX']     = conf.find_program('ocamllex', var='OCAMLLEX', mandatory=False)
	v['OCAMLYACC']    = conf.find_program('ocamlyacc', var='OCAMLYACC', mandatory=False)
	v['OCAMLFLAGS']   = ''
	where = conf.cmd_and_log(conf.env.OCAMLC + ['-where']).strip()+os.sep
	v['OCAMLLIB']     = where
	v['LIBPATH_OCAML'] = where
	v['INCLUDES_OCAML'] = where
	v['LIB_OCAML'] = 'camlrun'




#!/usr/bin/env python
# encoding: utf-8
# Rafaël Kooi 2019

from waflib import TaskGen

@TaskGen.feature('c', 'cxx', 'fc')
@TaskGen.after_method('propagate_uselib_vars')
def add_pdb_per_object(self):
	"""For msvc/fortran, specify a unique compile pdb per object, to work
	around LNK4099. Flags are updated with a unique /Fd flag based on the
	task output name. This is separate from the link pdb.
	"""
	if not hasattr(self, 'compiled_tasks'):
		return

	link_task = getattr(self, 'link_task', None)

	for task in self.compiled_tasks:
		if task.inputs and task.inputs[0].name.lower().endswith('.rc'):
			continue

		add_pdb = False
		for flagname in ('CFLAGS', 'CXXFLAGS', 'FCFLAGS'):
			# several languages may be used at once
			for flag in task.env[flagname]:
				if flag[1:].lower() == 'zi':
					add_pdb = True
					break

		if add_pdb:
			node = task.outputs[0].change_ext('.pdb')
			pdb_flag = '/Fd:' + node.abspath()

			for flagname in ('CFLAGS', 'CXXFLAGS', 'FCFLAGS'):
				buf = [pdb_flag]
				for flag in task.env[flagname]:
					if flag[1:3] == 'Fd' or flag[1:].lower() == 'fs' or flag[1:].lower() == 'mp':
						continue
					buf.append(flag)
				task.env[flagname] = buf

			if link_task and not node in link_task.dep_nodes:
				link_task.dep_nodes.append(node)
			if not node in task.outputs:
				task.outputs.append(node)



#! /usr/bin/env python
# encoding: utf-8
# Detection of the flang Fortran compiler

import re, shlex
from waflib.Tools import fc,fc_config,fc_scan
from waflib.Configure import conf
from waflib import Logs, Errors, Utils
from waflib.Tools.compiler_fc import fc_compiler
fc_compiler['linux'].append('fc_flang')

from waflib.Tools.c_config import MACRO_TO_DESTOS, MACRO_TO_DEST_CPU

@conf
def find_flang(conf):
	fc = conf.find_program(['flang'], var='FC')
	conf.get_nfort_version(fc)
	conf.env.FC_NAME = 'FLANG'
	conf.env.FC_MOD_CAPITALIZATION = 'lower'

@conf
def flang_flags(conf):
	v = conf.env
	v['_FCMODOUTFLAGS'] = []
	v['FCFLAGS_DEBUG'] = []
	v['FCFLAGS_fcshlib'] = []
	v['LINKFLAGS_fcshlib'] = []
	v['FCSTLIB_MARKER'] = ''
	v['FCSHLIB_MARKER'] = ''

@conf
def get_flang_version(conf, fc):
	cmd = fc + ['-dM', '-E', '-']
	env = conf.env.env or None

	try:
		out, err = conf.cmd_and_log(cmd, output=0, input='\n'.encode(), env=env)
	except Errors.WafError:
		conf.fatal('Could not determine the FLANG compiler version for %r' % cmd)
	if out.find('__clang__') < 0:
		conf.fatal('Not a flang compiler')

	k = {}
	out = out.splitlines()
	for line in out:
		lst = shlex.split(line)
		if len(lst)>2:
			key = lst[1]
			val = lst[2]
			k[key] = val

	def isD(var):
		return var in k

	# Some documentation is available at http://predef.sourceforge.net
	# The names given to DEST_OS must match what Utils.unversioned_sys_platform() returns.
	if not conf.env.DEST_OS:
		conf.env.DEST_OS = ''
	for i in MACRO_TO_DESTOS:
		if isD(i):
			conf.env.DEST_OS = MACRO_TO_DESTOS[i]
			break
	else:
		if isD('__APPLE__') and isD('__MACH__'):
			conf.env.DEST_OS = 'darwin'
		elif isD('__unix__'): # unix must be tested last as it's a generic fallback
			conf.env.DEST_OS = 'generic'

	if isD('__ELF__'):
		conf.env.DEST_BINFMT = 'elf'
	elif isD('__WINNT__') or isD('__CYGWIN__') or isD('_WIN32'):
		conf.env.DEST_BINFMT = 'pe'
		if not conf.env.IMPLIBDIR:
			conf.env.IMPLIBDIR = conf.env.LIBDIR # for .lib or .dll.a files
		conf.env.LIBDIR = conf.env.BINDIR
	elif isD('__APPLE__'):
		conf.env.DEST_BINFMT = 'mac-o'

	if not conf.env.DEST_BINFMT:
		# Infer the binary format from the os name.
		conf.env.DEST_BINFMT = Utils.destos_to_binfmt(conf.env.DEST_OS)

	for i in MACRO_TO_DEST_CPU:
		if isD(i):
			conf.env.DEST_CPU = MACRO_TO_DEST_CPU[i]
			break

	Logs.debug('fc_flang: dest platform: ' + ' '.join([conf.env[x] or '?' for x in ('DEST_OS', 'DEST_BINFMT', 'DEST_CPU')]))
	conf.env.FC_VERSION = (k['__clang_major__'], k['__clang_minor__'], k['__clang_patchlevel__'])

	return k


def configure(conf):
	conf.find_flang()
	conf.find_ar()
	conf.fc_flags()
	conf.fc_add_flags()
	conf.flang_flags()



#! /usr/bin/env python
# encoding: UTF-8
# Thomas Nagy 2008-2010 (ita)

"""

Doxygen support

Variables passed to bld():
* doxyfile -- the Doxyfile to use
* doxy_tar -- destination archive for generated documentation (if desired)
* install_path -- where to install the documentation
* pars -- dictionary overriding doxygen configuration settings

When using this tool, the wscript will look like:

	def options(opt):
		opt.load('doxygen')

	def configure(conf):
		conf.load('doxygen')
		# check conf.env.DOXYGEN, if it is mandatory

	def build(bld):
		if bld.env.DOXYGEN:
			bld(features="doxygen", doxyfile='Doxyfile', ...)
"""

import os, os.path, re
from collections import OrderedDict
from waflib import Task, Utils, Node
from waflib.TaskGen import feature

DOXY_STR = '"${DOXYGEN}" - '
DOXY_FMTS = 'html latex man rft xml'.split()
DOXY_FILE_PATTERNS = '*.' + ' *.'.join('''
c cc cxx cpp c++ java ii ixx ipp i++ inl h hh hxx hpp h++ idl odl cs php php3
inc m mm py f90c cc cxx cpp c++ java ii ixx ipp i++ inl h hh hxx
'''.split())

re_rl = re.compile('\\\\\r*\n', re.MULTILINE)
re_nl = re.compile('\r*\n', re.M)
def parse_doxy(txt):
	'''
	Parses a doxygen file.
	Returns an ordered dictionary. We cannot return a default dictionary, as the
	order in which the entries are reported does matter, especially for the
	'@INCLUDE' lines.
	'''
	tbl = OrderedDict()
	txt   = re_rl.sub('', txt)
	lines = re_nl.split(txt)
	for x in lines:
		x = x.strip()
		if not x or x.startswith('#') or x.find('=') < 0:
			continue
		if x.find('+=') >= 0:
			tmp = x.split('+=')
			key = tmp[0].strip()
			if key in tbl:
				tbl[key] += ' ' + '+='.join(tmp[1:]).strip()
			else:
				tbl[key] = '+='.join(tmp[1:]).strip()
		else:
			tmp = x.split('=')
			tbl[tmp[0].strip()] = '='.join(tmp[1:]).strip()
	return tbl

class doxygen(Task.Task):
	vars  = ['DOXYGEN', 'DOXYFLAGS']
	color = 'BLUE'
	ext_in = [ '.py', '.c', '.h', '.java', '.pb.cc' ]

	def runnable_status(self):
		'''
		self.pars are populated in runnable_status - because this function is being
		run *before* both self.pars "consumers" - scan() and run()

		set output_dir (node) for the output
		'''

		for x in self.run_after:
			if not x.hasrun:
				return Task.ASK_LATER

		if not getattr(self, 'pars', None):
			txt = self.inputs[0].read()
			self.pars = parse_doxy(txt)

			# Override with any parameters passed to the task generator
			if getattr(self.generator, 'pars', None):
				for k, v in self.generator.pars.items():
					self.pars[k] = v

			if self.pars.get('OUTPUT_DIRECTORY'):
				# Use the path parsed from the Doxyfile as an absolute path
				output_node = self.inputs[0].parent.get_bld().make_node(self.pars['OUTPUT_DIRECTORY'])
			else:
				# If no OUTPUT_PATH was specified in the Doxyfile, build path from the Doxyfile name + '.doxy'
				output_node = self.inputs[0].parent.get_bld().make_node(self.inputs[0].name + '.doxy')
			output_node.mkdir()
			self.pars['OUTPUT_DIRECTORY'] = output_node.abspath()

			self.doxy_inputs = getattr(self, 'doxy_inputs', [])
			if not self.pars.get('INPUT'):
				self.doxy_inputs.append(self.inputs[0].parent)
			else:
				for i in self.pars.get('INPUT').split():
					if os.path.isabs(i):
						node = self.generator.bld.root.find_node(i)
					else:
						node = self.inputs[0].parent.find_node(i)
					if not node:
						self.generator.bld.fatal('Could not find the doxygen input %r' % i)
					self.doxy_inputs.append(node)

		if not getattr(self, 'output_dir', None):
			bld = self.generator.bld
			# Output path is always an absolute path as it was transformed above.
			self.output_dir = bld.root.find_dir(self.pars['OUTPUT_DIRECTORY'])

		self.signature()
		ret = Task.Task.runnable_status(self)
		if ret == Task.SKIP_ME:
			# in case the files were removed
			self.add_install()
		return ret

	def scan(self):
		exclude_patterns = self.pars.get('EXCLUDE_PATTERNS','').split()
		exclude_patterns = [pattern.replace('*/', '**/') for pattern in exclude_patterns]
		file_patterns = self.pars.get('FILE_PATTERNS','').split()
		if not file_patterns:
			file_patterns = DOXY_FILE_PATTERNS.split()
		if self.pars.get('RECURSIVE') == 'YES':
			file_patterns = ["**/%s" % pattern for pattern in file_patterns]
		nodes = []
		names = []
		for node in self.doxy_inputs:
			if os.path.isdir(node.abspath()):
				for m in node.ant_glob(incl=file_patterns, excl=exclude_patterns):
					nodes.append(m)
			else:
				nodes.append(node)
		return (nodes, names)

	def run(self):
		dct = self.pars.copy()
		code = '\n'.join(['%s = %s' % (x, dct[x]) for x in self.pars])
		code = code.encode() # for python 3
		#fmt = DOXY_STR % (self.inputs[0].parent.abspath())
		cmd = Utils.subst_vars(DOXY_STR, self.env)
		env = self.env.env or None
		proc = Utils.subprocess.Popen(cmd, shell=True, stdin=Utils.subprocess.PIPE, env=env, cwd=self.inputs[0].parent.abspath())
		proc.communicate(code)
		return proc.returncode

	def post_run(self):
		nodes = self.output_dir.ant_glob('**/*', quiet=True)
		for x in nodes:
			self.generator.bld.node_sigs[x] = self.uid()
		self.add_install()
		return Task.Task.post_run(self)

	def add_install(self):
		nodes = self.output_dir.ant_glob('**/*', quiet=True)
		self.outputs += nodes
		if getattr(self.generator, 'install_path', None):
			if not getattr(self.generator, 'doxy_tar', None):
				self.generator.add_install_files(install_to=self.generator.install_path,
					install_from=self.outputs,
					postpone=False,
					cwd=self.output_dir,
					relative_trick=True)

class tar(Task.Task):
	"quick tar creation"
	run_str = '${TAR} ${TAROPTS} ${TGT} ${SRC}'
	color   = 'RED'
	after   = ['doxygen']
	def runnable_status(self):
		for x in getattr(self, 'input_tasks', []):
			if not x.hasrun:
				return Task.ASK_LATER

		if not getattr(self, 'tar_done_adding', None):
			# execute this only once
			self.tar_done_adding = True
			for x in getattr(self, 'input_tasks', []):
				self.set_inputs(x.outputs)
			if not self.inputs:
				return Task.SKIP_ME
		return Task.Task.runnable_status(self)

	def __str__(self):
		tgt_str = ' '.join([a.path_from(a.ctx.launch_node()) for a in self.outputs])
		return '%s: %s\n' % (self.__class__.__name__, tgt_str)

@feature('doxygen')
def process_doxy(self):
	if not getattr(self, 'doxyfile', None):
		self.bld.fatal('no doxyfile variable specified??')

	node = self.doxyfile
	if not isinstance(node, Node.Node):
		node = self.path.find_resource(node)
	if not node:
		self.bld.fatal('doxygen file %s not found' % self.doxyfile)

	# the task instance
	dsk = self.create_task('doxygen', node, always_run=getattr(self, 'always', False))

	if getattr(self, 'doxy_tar', None):
		tsk = self.create_task('tar', always_run=getattr(self, 'always', False))
		tsk.input_tasks = [dsk]
		tsk.set_outputs(self.path.find_or_declare(self.doxy_tar))
		if self.doxy_tar.endswith('bz2'):
			tsk.env['TAROPTS'] = ['cjf']
		elif self.doxy_tar.endswith('gz'):
			tsk.env['TAROPTS'] = ['czf']
		else:
			tsk.env['TAROPTS'] = ['cf']
		if getattr(self, 'install_path', None):
			self.add_install_files(install_to=self.install_path, install_from=tsk.outputs)

def configure(conf):
	'''
	Check if doxygen and tar commands are present in the system

	If the commands are present, then conf.env.DOXYGEN and conf.env.TAR
	variables will be set. Detection can be controlled by setting DOXYGEN and
	TAR environmental variables.
	'''

	conf.find_program('doxygen', var='DOXYGEN', mandatory=False)
	conf.find_program('tar', var='TAR', mandatory=False)



#! /usr/bin/env python
# encoding: utf-8

"""
Use extended attributes instead of database files

1. Input files will be made writable
2. This is only for systems providing extended filesystem attributes
3. By default, hashes are calculated only if timestamp/size change (HASH_CACHE below)
4. The module enables "deep_inputs" on all tasks by propagating task signatures
5. This module also skips task signature comparisons for task code changes due to point 4.
6. This module is for Python3/Linux only, but it could be extended to Python2/other systems
   using the xattr library
7. For projects in which tasks always declare output files, it should be possible to
   store the rest of build context attributes on output files (imp_sigs, raw_deps and node_deps)
   but this is not done here

On a simple C++ project benchmark, the variations before and after adding waf_xattr.py were observed:
total build time: 20s -> 22s
no-op build time: 2.4s -> 1.8s
pickle file size: 2.9MB -> 2.6MB
"""

import os
from waflib import Logs, Node, Task, Utils, Errors
from waflib.Task import SKIP_ME, RUN_ME, CANCEL_ME, ASK_LATER, SKIPPED, MISSING

HASH_CACHE = True
SIG_VAR = 'user.waf.sig'
SEP = ','.encode()
TEMPLATE = '%b%d,%d'.encode()

try:
	PermissionError
except NameError:
	PermissionError = IOError

def getxattr(self):
	return os.getxattr(self.abspath(), SIG_VAR)

def setxattr(self, val):
	os.setxattr(self.abspath(), SIG_VAR, val)

def h_file(self):
	try:
		ret = getxattr(self)
	except OSError:
		if HASH_CACHE:
			st = os.stat(self.abspath())
			mtime = st.st_mtime
			size = st.st_size
	else:
		if len(ret) == 16:
			# for build directory files
			return ret

		if HASH_CACHE:
			# check if timestamp and mtime match to avoid re-hashing
			st = os.stat(self.abspath())
			mtime, size = ret[16:].split(SEP)
			if int(1000 * st.st_mtime) == int(mtime) and st.st_size == int(size):
				return ret[:16]

	ret = Utils.h_file(self.abspath())
	if HASH_CACHE:
		val = TEMPLATE % (ret, int(1000 * st.st_mtime), int(st.st_size))
		try:
			setxattr(self, val)
		except PermissionError:
			os.chmod(self.abspath(), st.st_mode | 128)
			setxattr(self, val)
	return ret

def runnable_status(self):
	bld = self.generator.bld
	if bld.is_install < 0:
		return SKIP_ME

	for t in self.run_after:
		if not t.hasrun:
			return ASK_LATER
		elif t.hasrun < SKIPPED:
			# a dependency has an error
			return CANCEL_ME

	# first compute the signature
	try:
		new_sig = self.signature()
	except Errors.TaskNotReady:
		return ASK_LATER

	if not self.outputs:
		# compare the signature to a signature computed previously
		# this part is only for tasks with no output files
		key = self.uid()
		try:
			prev_sig = bld.task_sigs[key]
		except KeyError:
			Logs.debug('task: task %r must run: it was never run before or the task code changed', self)
			return RUN_ME
		if new_sig != prev_sig:
			Logs.debug('task: task %r must run: the task signature changed', self)
			return RUN_ME

	# compare the signatures of the outputs to make a decision
	for node in self.outputs:
		try:
			sig = node.h_file()
		except EnvironmentError:
			Logs.debug('task: task %r must run: an output node does not exist', self)
			return RUN_ME
		if sig != new_sig:
			Logs.debug('task: task %r must run: an output node is stale', self)
			return RUN_ME

	return (self.always_run and RUN_ME) or SKIP_ME

def post_run(self):
	bld = self.generator.bld
	sig = self.signature()
	for node in self.outputs:
		if not node.exists():
			self.hasrun = MISSING
			self.err_msg = '-> missing file: %r' % node.abspath()
			raise Errors.WafError(self.err_msg)
		os.setxattr(node.abspath(), 'user.waf.sig', sig)
	if not self.outputs:
		# only for task with no outputs
		bld.task_sigs[self.uid()] = sig
	if not self.keep_last_cmd:
		try:
			del self.last_cmd
		except AttributeError:
			pass

try:
	os.getxattr
except AttributeError:
	pass
else:
	h_file.__doc__ = Node.Node.h_file.__doc__

	# keep file hashes as file attributes
	Node.Node.h_file = h_file

	# enable "deep_inputs" on all tasks
	Task.Task.runnable_status = runnable_status
	Task.Task.post_run = post_run
	Task.Task.sig_deep_inputs = Utils.nada




#!/usr/bin/env python
# encoding: utf-8

# Replaces the default formatter by one which understands MSVC output and colorizes it.
# Modified from color_gcc.py

__author__ = __maintainer__ = "Alibek Omarov <a1ba.omarov@gmail.com>"
__copyright__ = "Alibek Omarov, 2019"

import sys
from waflib import Logs

class ColorMSVCFormatter(Logs.formatter):
	def __init__(self, colors):
		self.colors = colors
		Logs.formatter.__init__(self)

	def parseMessage(self, line, color):
		# Split messaage from 'disk:filepath: type: message'
		arr = line.split(':', 3)
		if len(arr) < 4:
			return line

		colored = self.colors.BOLD + arr[0] + ':' + arr[1] + ':' + self.colors.NORMAL
		colored += color + arr[2] + ':' + self.colors.NORMAL
		colored += arr[3]
		return colored

	def format(self, rec):
		frame = sys._getframe()
		while frame:
			func = frame.f_code.co_name
			if func == 'exec_command':
				cmd = frame.f_locals.get('cmd')
				if isinstance(cmd, list):
					# Fix file case, it may be CL.EXE or cl.exe
					argv0 = cmd[0].lower()
					if 'cl.exe' in argv0:
						lines = []
						# This will not work with "localized" versions
						# of MSVC
						for line in rec.msg.splitlines():
							if ': warning ' in line:
								lines.append(self.parseMessage(line, self.colors.YELLOW))
							elif ': error ' in line:
								lines.append(self.parseMessage(line, self.colors.RED))
							elif ': fatal error ' in line:
								lines.append(self.parseMessage(line, self.colors.RED + self.colors.BOLD))
							elif ': note: ' in line:
								lines.append(self.parseMessage(line, self.colors.CYAN))
							else:
								lines.append(line)
						rec.msg = "\n".join(lines)
			frame = frame.f_back
		return Logs.formatter.format(self, rec)

def options(opt):
	Logs.log.handlers[0].setFormatter(ColorMSVCFormatter(Logs.colors))




#!/usr/bin/python
# encoding: utf-8
# vim: tabstop=4 noexpandtab

"""
Create a satellite assembly from "*.??.txt" files. ?? stands for a language code.

The projects Resources subfolder contains resources.??.txt string files for several languages.
The build folder will hold the satellite assemblies as ./??/ExeName.resources.dll

#gen becomes template (It is called gen because it also uses resx.py).
bld(source='Resources/resources.de.txt',gen=ExeName)
"""

import os, re
from waflib import Task
from waflib.TaskGen import feature,before_method

class al(Task.Task):
	run_str = '${AL} ${ALFLAGS}'

@feature('satellite_assembly')
@before_method('process_source')
def satellite_assembly(self):
	if not getattr(self, 'gen', None):
		self.bld.fatal('satellite_assembly needs a template assembly provided with the "gen" parameter')
	res_lang = re.compile(r'(.*)\.(\w\w)\.(?:resx|txt)',flags=re.I)

	# self.source can contain node objects, so this will break in one way or another
	self.source = self.to_list(self.source)
	for i, x in enumerate(self.source):
		#x = 'resources/resources.de.resx'
		#x = 'resources/resources.de.txt'
		mo = res_lang.match(x)
		if mo:
			template = os.path.splitext(self.gen)[0]
			templatedir, templatename = os.path.split(template)
			res = mo.group(1)
			lang = mo.group(2)
			#./Resources/resources.de.resources
			resources = self.path.find_or_declare(res+ '.' + lang + '.resources')
			self.create_task('resgen', self.to_nodes(x), [resources])
			#./de/Exename.resources.dll
			satellite = self.path.find_or_declare(os.path.join(templatedir,lang,templatename) + '.resources.dll')
			tsk = self.create_task('al',[resources],[satellite])
			tsk.env.append_value('ALFLAGS','/template:'+os.path.join(self.path.relpath(),self.gen))
			tsk.env.append_value('ALFLAGS','/embed:'+resources.relpath())
			tsk.env.append_value('ALFLAGS','/culture:'+lang)
			tsk.env.append_value('ALFLAGS','/out:'+satellite.relpath())
			self.source[i] = None
	# remove the None elements that we just substituted
	self.source = list(filter(lambda x:x, self.source))

def configure(ctx):
	ctx.find_program('al', var='AL', mandatory=True)
	ctx.load('resx')




#!/usr/bin/env python
# encoding: utf-8
# Carlos Rafael Giani, 2007 (dv)
# Thomas Nagy, 2007-2018 (ita)

from waflib import Utils, Task, Errors
from waflib.TaskGen import taskgen_method, feature, extension
from waflib.Tools import d_scan, d_config
from waflib.Tools.ccroot import link_task, stlink_task

class d(Task.Task):
	"Compile a d file into an object file"
	color   = 'GREEN'
	run_str = '${D} ${DFLAGS} ${DINC_ST:INCPATHS} ${D_SRC_F:SRC} ${D_TGT_F:TGT}'
	scan    = d_scan.scan

class d_with_header(d):
	"Compile a d file and generate a header"
	run_str = '${D} ${DFLAGS} ${DINC_ST:INCPATHS} ${D_HDR_F:tgt.outputs[1].bldpath()} ${D_SRC_F:SRC} ${D_TGT_F:tgt.outputs[0].bldpath()}'

class d_header(Task.Task):
	"Compile d headers"
	color   = 'BLUE'
	run_str = '${D} ${D_HEADER} ${SRC}'

class dprogram(link_task):
	"Link object files into a d program"
	run_str = '${D_LINKER} ${LINKFLAGS} ${DLNK_SRC_F}${SRC} ${DLNK_TGT_F:TGT} ${RPATH_ST:RPATH} ${DSTLIB_MARKER} ${DSTLIBPATH_ST:STLIBPATH} ${DSTLIB_ST:STLIB} ${DSHLIB_MARKER} ${DLIBPATH_ST:LIBPATH} ${DSHLIB_ST:LIB}'
	inst_to = '${BINDIR}'

class dshlib(dprogram):
	"Link object files into a d shared library"
	inst_to = '${LIBDIR}'

class dstlib(stlink_task):
	"Link object files into a d static library"
	pass # do not remove

@extension('.d', '.di', '.D')
def d_hook(self, node):
	"""
	Compile *D* files. To get .di files as well as .o files, set the following::

		def build(bld):
			bld.program(source='foo.d', target='app', generate_headers=True)

	"""
	ext = Utils.destos_to_binfmt(self.env.DEST_OS) == 'pe' and 'obj' or 'o'
	out = '%s.%d.%s' % (node.name, self.idx, ext)
	def create_compiled_task(self, name, node):
		task = self.create_task(name, node, node.parent.find_or_declare(out))
		try:
			self.compiled_tasks.append(task)
		except AttributeError:
			self.compiled_tasks = [task]
		return task

	if getattr(self, 'generate_headers', None):
		tsk = create_compiled_task(self, 'd_with_header', node)
		tsk.outputs.append(node.change_ext(self.env.DHEADER_ext))
	else:
		tsk = create_compiled_task(self, 'd', node)
	return tsk

@taskgen_method
def generate_header(self, filename):
	"""
	See feature request #104::

		def build(bld):
			tg = bld.program(source='foo.d', target='app')
			tg.generate_header('blah.d')
			# is equivalent to:
			#tg = bld.program(source='foo.d', target='app', header_lst='blah.d')

	:param filename: header to create
	:type filename: string
	"""
	try:
		self.header_lst.append([filename, self.install_path])
	except AttributeError:
		self.header_lst = [[filename, self.install_path]]

@feature('d')
def process_header(self):
	"""
	Process the attribute 'header_lst' to create the d header compilation tasks::

		def build(bld):
			bld.program(source='foo.d', target='app', header_lst='blah.d')
	"""
	for i in getattr(self, 'header_lst', []):
		node = self.path.find_resource(i[0])
		if not node:
			raise Errors.WafError('file %r not found on d obj' % i[0])
		self.create_task('d_header', node, node.change_ext('.di'))




#!/usr/bin/env python
# encoding: utf-8
# Hans-Martin von Gaudecker, 2012

"""
Run a Stata do-script in the directory specified by **ctx.bldnode**. The
first and only argument will be the name of the do-script (no extension),
which can be accessed inside the do-script by the local macro `1'. Useful
for keeping a log file.

The tool uses the log file that is automatically kept by Stata only 
for error-catching purposes, it will be destroyed if the task finished
without error. In case of an error in **some_script.do**, you can inspect
it as **some_script.log** in the **ctx.bldnode** directory.

Note that Stata will not return an error code if it exits abnormally -- 
catching errors relies on parsing the log file mentioned before. Should
the parser behave incorrectly please send an email to hmgaudecker [at] gmail.

**WARNING**

    The tool will not work if multiple do-scripts of the same name---but in
    different directories---are run at the same time! Avoid this situation.

Usage::

    ctx(features='run_do_script', 
        source='some_script.do',
        target=['some_table.tex', 'some_figure.eps'],
        deps='some_data.csv')
"""


import os, re, sys
from waflib import Task, TaskGen, Logs

if sys.platform == 'darwin':
	STATA_COMMANDS = ['Stata64MP', 'StataMP',
								'Stata64SE', 'StataSE', 
								'Stata64', 'Stata']
	STATAFLAGS = '-e -q do'
	STATAENCODING = 'MacRoman'
elif sys.platform.startswith('linux'):
	STATA_COMMANDS = ['stata-mp', 'stata-se', 'stata']
	STATAFLAGS = '-b -q do'
	# Not sure whether this is correct...
	STATAENCODING = 'Latin-1'
elif sys.platform.lower().startswith('win'):
	STATA_COMMANDS = ['StataMP-64', 'StataMP-ia',
								'StataMP', 'StataSE-64',
								'StataSE-ia', 'StataSE',
								'Stata-64', 'Stata-ia',
								'Stata.e', 'WMPSTATA',
								'WSESTATA', 'WSTATA']
	STATAFLAGS = '/e do'
	STATAENCODING = 'Latin-1'
else:
	raise Exception("Unknown sys.platform: %s " % sys.platform)

def configure(ctx):
	ctx.find_program(STATA_COMMANDS, var='STATACMD', errmsg="""\n
No Stata executable found!\n\n
If Stata is needed:\n
	1) Check the settings of your system path.
	2) Note we are looking for Stata executables called: %s
	   If yours has a different name, please report to hmgaudecker [at] gmail\n
Else:\n
	Do not load the 'run_do_script' tool in the main wscript.\n\n""" % STATA_COMMANDS)
	ctx.env.STATAFLAGS = STATAFLAGS
	ctx.env.STATAENCODING = STATAENCODING

class run_do_script_base(Task.Task):
	"""Run a Stata do-script from the bldnode directory."""
	run_str = '"${STATACMD}" ${STATAFLAGS} "${SRC[0].abspath()}" "${DOFILETRUNK}"'
	shell = True

class run_do_script(run_do_script_base):
	"""Use the log file automatically kept by Stata for error-catching.
	Erase it if the task finished without error. If not, it will show 
	up as do_script.log in the bldnode directory.
	"""
	def run(self):
		run_do_script_base.run(self)
		ret, log_tail  = self.check_erase_log_file()
		if ret:
			Logs.error("""Running Stata on %r failed with code %r.\n\nCheck the log file %s, last 10 lines\n\n%s\n\n\n""",
				self.inputs[0], ret, self.env.LOGFILEPATH, log_tail)
		return ret

	def check_erase_log_file(self):
		"""Parse Stata's default log file and erase it if everything okay.

		Parser is based on Brendan Halpin's shell script found here:
			http://teaching.sociology.ul.ie/bhalpin/wordpress/?p=122
		"""

		if sys.version_info.major >= 3:
			kwargs = {'file': self.env.LOGFILEPATH, 'mode': 'r', 'encoding': self.env.STATAENCODING}
		else:
			kwargs = {'name': self.env.LOGFILEPATH, 'mode': 'r'}
		with open(**kwargs) as log:
			log_tail = log.readlines()[-10:]
			for line in log_tail:
				error_found = re.match(r"r\(([0-9]+)\)", line)
				if error_found:
					return error_found.group(1), ''.join(log_tail)
				else:
					pass
		# Only end up here if the parser did not identify an error.
		os.remove(self.env.LOGFILEPATH)
		return None, None


@TaskGen.feature('run_do_script')
@TaskGen.before_method('process_source')
def apply_run_do_script(tg):
	"""Task generator customising the options etc. to call Stata in batch
	mode for running a do-script.
	"""

	# Convert sources and targets to nodes
	src_node = tg.path.find_resource(tg.source)
	tgt_nodes = [tg.path.find_or_declare(t) for t in tg.to_list(tg.target)]

	tsk = tg.create_task('run_do_script', src=src_node, tgt=tgt_nodes)
	tsk.env.DOFILETRUNK = os.path.splitext(src_node.name)[0]
	tsk.env.LOGFILEPATH = os.path.join(tg.bld.bldnode.abspath(), '%s.log' % (tsk.env.DOFILETRUNK))

	# dependencies (if the attribute 'deps' changes, trigger a recompilation)
	for x in tg.to_list(getattr(tg, 'deps', [])):
		node = tg.path.find_resource(x)
		if not node:
			tg.bld.fatal('Could not find dependency %r for running %r' % (x, src_node.abspath()))
		tsk.dep_nodes.append(node)
	Logs.debug('deps: found dependencies %r for running %r', tsk.dep_nodes, src_node.abspath())

	# Bypass the execution of process_source by setting the source to an empty list
	tg.source = []




#! /usr/bin/env python
# encoding: utf-8
# Eclipse CDT 5.0 generator for Waf
# Richard Quirk 2009-1011 (New BSD License)
# Thomas Nagy 2011 (ported to Waf 1.6)

"""
Usage:

def options(opt):
	opt.load('eclipse')

To add additional targets beside standard ones (configure, dist, install, check)
the environment ECLIPSE_EXTRA_TARGETS can be set (ie. to ['test', 'lint', 'docs'])

$ waf configure eclipse
"""

import sys, os
from waflib import Utils, Logs, Context, Build, TaskGen, Scripting, Errors, Node
from xml.dom.minidom import Document

STANDARD_INCLUDES = [ '/usr/local/include', '/usr/include' ]

oe_cdt = 'org.eclipse.cdt'
cdt_mk = oe_cdt + '.make.core'
cdt_core = oe_cdt + '.core'
cdt_bld = oe_cdt + '.build.core'
extbuilder_dir = '.externalToolBuilders'
extbuilder_name = 'Waf_Builder.launch'
settings_dir = '.settings'
settings_name = 'language.settings.xml'

class eclipse(Build.BuildContext):
	cmd = 'eclipse'
	fun = Scripting.default_cmd

	def execute(self):
		"""
		Entry point
		"""
		self.restore()
		if not self.all_envs:
			self.load_envs()
		self.recurse([self.run_dir])

		appname = getattr(Context.g_module, Context.APPNAME, os.path.basename(self.srcnode.abspath()))
		self.create_cproject(appname, pythonpath=self.env['ECLIPSE_PYTHON_PATH'])

	# Helper to dump the XML document content to XML with UTF-8 encoding
	def write_conf_to_xml(self, filename, document):
		self.srcnode.make_node(filename).write(document.toprettyxml(encoding='UTF-8'), flags='wb')

	def create_cproject(self, appname, workspace_includes=[], pythonpath=[]):
		"""
		Create the Eclipse CDT .project and .cproject files
		@param appname The name that will appear in the Project Explorer
		@param build The BuildContext object to extract includes from
		@param workspace_includes Optional project includes to prevent
			  "Unresolved Inclusion" errors in the Eclipse editor
		@param pythonpath Optional project specific python paths
		"""
		hasc = hasjava = haspython = False
		source_dirs = []
		cpppath = self.env['CPPPATH']
		javasrcpath = []
		javalibpath = []
		includes = STANDARD_INCLUDES
		if sys.platform != 'win32':
			cc = self.env.CC or self.env.CXX
			if cc:
				cmd = cc + ['-xc++', '-E', '-Wp,-v', '-']
				try:
					gccout = self.cmd_and_log(cmd, output=Context.STDERR, quiet=Context.BOTH, input='\n'.encode()).splitlines()
				except Errors.WafError:
					pass
				else:
					includes = []
					for ipath in gccout:
						if ipath.startswith(' /'):
							includes.append(ipath[1:])
			cpppath += includes
		Logs.warn('Generating Eclipse CDT project files')

		for g in self.groups:
			for tg in g:
				if not isinstance(tg, TaskGen.task_gen):
					continue

				tg.post()

				# Add local Python modules paths to configuration so object resolving will work in IDE
				# This may also contain generated files (ie. pyqt5 or protoc) that get picked from build
				if 'py' in tg.features:
					pypath = tg.path.relpath()
					py_installfrom = getattr(tg, 'install_from', None)
					if isinstance(py_installfrom, Node.Node):
						pypath = py_installfrom.path_from(self.root.make_node(self.top_dir))
					if pypath not in pythonpath:
						pythonpath.append(pypath)
					haspython = True

				# Add Java source directories so object resolving works in IDE
				# This may also contain generated files (ie. protoc) that get picked from build
				if 'javac' in tg.features:
					java_src = tg.path.relpath()
					java_srcdir = getattr(tg.javac_task, 'srcdir', None)
					if java_srcdir:
						if isinstance(java_srcdir, Node.Node):
							java_srcdir = [java_srcdir]
						for x in Utils.to_list(java_srcdir):
							x = x.path_from(self.root.make_node(self.top_dir))
							if x not in javasrcpath:
								javasrcpath.append(x)
					else:
						if java_src not in javasrcpath:
							javasrcpath.append(java_src)
					hasjava = True

					# Check if there are external dependencies and add them as external jar so they will be resolved by Eclipse
					usedlibs=getattr(tg, 'use', [])
					for x in Utils.to_list(usedlibs):
						for cl in Utils.to_list(tg.env['CLASSPATH_'+x]):
							if cl not in javalibpath:
								javalibpath.append(cl)

				if not getattr(tg, 'link_task', None):
					continue

				features = Utils.to_list(getattr(tg, 'features', ''))

				is_cc = 'c' in features or 'cxx' in features

				incnodes = tg.to_incnodes(tg.to_list(getattr(tg, 'includes', [])) + tg.env['INCLUDES'])
				for p in incnodes:
					path = p.path_from(self.srcnode)

					if (path.startswith("/")):
						if path not in cpppath:
							cpppath.append(path)
					else:
						if path not in workspace_includes:
							workspace_includes.append(path)

					if is_cc and path not in source_dirs:
						source_dirs.append(path)

					hasc = True

		waf_executable = os.path.abspath(sys.argv[0])
		project = self.impl_create_project(sys.executable, appname, hasc, hasjava, haspython, waf_executable)
		self.write_conf_to_xml('.project', project)

		if hasc:
			project = self.impl_create_cproject(sys.executable, waf_executable, appname, workspace_includes, cpppath, source_dirs)
			self.write_conf_to_xml('.cproject', project)

		if haspython:
			project = self.impl_create_pydevproject(sys.path, pythonpath)
			self.write_conf_to_xml('.pydevproject', project)

		if hasjava:
			project = self.impl_create_javaproject(javasrcpath, javalibpath)
			self.write_conf_to_xml('.classpath', project)

		# Create editor language settings to have correct standards applied in IDE, as per project configuration
		try:
			os.mkdir(settings_dir)
		except OSError:
			pass	# Ignore if dir already exists

		lang_settings = Document()
		project = lang_settings.createElement('project')

		# Language configurations for C and C++ via cdt
		if hasc:
			configuration = self.add(lang_settings, project, 'configuration',
							{'id' : 'org.eclipse.cdt.core.default.config.1', 'name': 'Default'})

			extension = self.add(lang_settings, configuration, 'extension', {'point': 'org.eclipse.cdt.core.LanguageSettingsProvider'})

			provider = self.add(lang_settings, extension, 'provider',
							{ 'copy-of': 'extension',
							  'id': 'org.eclipse.cdt.ui.UserLanguageSettingsProvider'})

			provider = self.add(lang_settings, extension, 'provider-reference',
							{ 'id': 'org.eclipse.cdt.core.ReferencedProjectsLanguageSettingsProvider',
							  'ref': 'shared-provider'})

			provider = self.add(lang_settings, extension, 'provider-reference',
							{ 'id': 'org.eclipse.cdt.managedbuilder.core.MBSLanguageSettingsProvider',
							  'ref': 'shared-provider'})

			# C and C++ are kept as separated providers so appropriate flags are used also in mixed projects
			if self.env.CC:
				provider = self.add(lang_settings, extension, 'provider',
							{ 'class': 'org.eclipse.cdt.managedbuilder.language.settings.providers.GCCBuiltinSpecsDetector',
							  'console': 'false',
							  'id': 'org.eclipse.cdt.managedbuilder.language.settings.providers.GCCBuiltinSpecsDetector.1',
							  'keep-relative-paths' : 'false',
							  'name': 'CDT GCC Built-in Compiler Settings',
							  'parameter': '%s %s ${FLAGS} -E -P -v -dD "${INPUTS}"'%(self.env.CC[0],' '.join(self.env['CFLAGS'])),
							  'prefer-non-shared': 'true' })

				self.add(lang_settings, provider, 'language-scope', { 'id': 'org.eclipse.cdt.core.gcc'})

			if self.env.CXX:
				provider = self.add(lang_settings, extension, 'provider',
							{ 'class': 'org.eclipse.cdt.managedbuilder.language.settings.providers.GCCBuiltinSpecsDetector',
							  'console': 'false',
							  'id': 'org.eclipse.cdt.managedbuilder.language.settings.providers.GCCBuiltinSpecsDetector.2',
							  'keep-relative-paths' : 'false',
							  'name': 'CDT GCC Built-in Compiler Settings',
							  'parameter': '%s %s ${FLAGS} -E -P -v -dD "${INPUTS}"'%(self.env.CXX[0],' '.join(self.env['CXXFLAGS'])),
							  'prefer-non-shared': 'true' })
				self.add(lang_settings, provider, 'language-scope', { 'id': 'org.eclipse.cdt.core.g++'})

		lang_settings.appendChild(project)
		self.write_conf_to_xml('%s%s%s'%(settings_dir, os.path.sep, settings_name), lang_settings)

	def impl_create_project(self, executable, appname, hasc, hasjava, haspython, waf_executable):
		doc = Document()
		projectDescription = doc.createElement('projectDescription')
		self.add(doc, projectDescription, 'name', appname)
		self.add(doc, projectDescription, 'comment')
		self.add(doc, projectDescription, 'projects')
		buildSpec = self.add(doc, projectDescription, 'buildSpec')
		buildCommand = self.add(doc, buildSpec, 'buildCommand')
		self.add(doc, buildCommand, 'triggers', 'clean,full,incremental,')
		arguments = self.add(doc, buildCommand, 'arguments')
		dictionaries = {}

		# If CDT is present, instruct this one to call waf as it is more flexible (separate build/clean ...)
		if hasc:
			self.add(doc, buildCommand, 'name', oe_cdt + '.managedbuilder.core.genmakebuilder')
			# the default make-style targets are overwritten by the .cproject values
			dictionaries = {
					cdt_mk + '.contents': cdt_mk + '.activeConfigSettings',
					cdt_mk + '.enableAutoBuild': 'false',
					cdt_mk + '.enableCleanBuild': 'true',
					cdt_mk + '.enableFullBuild': 'true',
					}
		else:
			# Otherwise for Java/Python an external builder tool is created that will call waf build
			self.add(doc, buildCommand, 'name', 'org.eclipse.ui.externaltools.ExternalToolBuilder')
			dictionaries = {
					'LaunchConfigHandle': '<project>/%s/%s'%(extbuilder_dir, extbuilder_name),
					}
			# The definition is in a separate directory XML file
			try:
				os.mkdir(extbuilder_dir)
			except OSError:
				pass	# Ignore error if already exists

			# Populate here the external builder XML calling waf
			builder = Document()
			launchConfiguration = doc.createElement('launchConfiguration')
			launchConfiguration.setAttribute('type', 'org.eclipse.ui.externaltools.ProgramBuilderLaunchConfigurationType')
			self.add(doc, launchConfiguration, 'booleanAttribute', {'key': 'org.eclipse.debug.ui.ATTR_LAUNCH_IN_BACKGROUND', 'value': 'false'})
			self.add(doc, launchConfiguration, 'booleanAttribute', {'key': 'org.eclipse.ui.externaltools.ATTR_TRIGGERS_CONFIGURED', 'value': 'true'})
			self.add(doc, launchConfiguration, 'stringAttribute', {'key': 'org.eclipse.ui.externaltools.ATTR_LOCATION', 'value': waf_executable})
			self.add(doc, launchConfiguration, 'stringAttribute', {'key': 'org.eclipse.ui.externaltools.ATTR_RUN_BUILD_KINDS', 'value': 'full,incremental,'})
			self.add(doc, launchConfiguration, 'stringAttribute', {'key': 'org.eclipse.ui.externaltools.ATTR_TOOL_ARGUMENTS', 'value': 'build'})
			self.add(doc, launchConfiguration, 'stringAttribute', {'key': 'org.eclipse.ui.externaltools.ATTR_WORKING_DIRECTORY', 'value': '${project_loc}'})
			builder.appendChild(launchConfiguration)
			# And write the XML to the file references before
			self.write_conf_to_xml('%s%s%s'%(extbuilder_dir, os.path.sep, extbuilder_name), builder)


		for k, v in dictionaries.items():
			self.addDictionary(doc, arguments, k, v)

		natures = self.add(doc, projectDescription, 'natures')

		if hasc:
			nature_list = """
				core.ccnature
				managedbuilder.core.ScannerConfigNature
				managedbuilder.core.managedBuildNature
				core.cnature
			""".split()
			for n in nature_list:
				self.add(doc, natures, 'nature', oe_cdt + '.' + n)

		if haspython:
			self.add(doc, natures, 'nature', 'org.python.pydev.pythonNature')
		if hasjava:
			self.add(doc, natures, 'nature', 'org.eclipse.jdt.core.javanature')

		doc.appendChild(projectDescription)
		return doc

	def impl_create_cproject(self, executable, waf_executable, appname, workspace_includes, cpppath, source_dirs=[]):
		doc = Document()
		doc.appendChild(doc.createProcessingInstruction('fileVersion', '4.0.0'))
		cconf_id = cdt_core + '.default.config.1'
		cproject = doc.createElement('cproject')
		storageModule = self.add(doc, cproject, 'storageModule',
				{'moduleId': cdt_core + '.settings'})
		cconf = self.add(doc, storageModule, 'cconfiguration', {'id':cconf_id})

		storageModule = self.add(doc, cconf, 'storageModule',
				{'buildSystemId': oe_cdt + '.managedbuilder.core.configurationDataProvider',
				 'id': cconf_id,
				 'moduleId': cdt_core + '.settings',
				 'name': 'Default'})

		self.add(doc, storageModule, 'externalSettings')

		extensions = self.add(doc, storageModule, 'extensions')
		extension_list = """
			VCErrorParser
			MakeErrorParser
			GCCErrorParser
			GASErrorParser
			GLDErrorParser
		""".split()
		self.add(doc, extensions, 'extension', {'id': cdt_core + '.ELF', 'point':cdt_core + '.BinaryParser'})
		for e in extension_list:
			self.add(doc, extensions, 'extension', {'id': cdt_core + '.' + e, 'point':cdt_core + '.ErrorParser'})

		storageModule = self.add(doc, cconf, 'storageModule',
				{'moduleId': 'cdtBuildSystem', 'version': '4.0.0'})
		config = self.add(doc, storageModule, 'configuration',
					{'artifactName': appname,
					 'id': cconf_id,
					 'name': 'Default',
					 'parent': cdt_bld + '.prefbase.cfg'})
		folderInfo = self.add(doc, config, 'folderInfo',
							{'id': cconf_id+'.', 'name': '/', 'resourcePath': ''})

		toolChain = self.add(doc, folderInfo, 'toolChain',
				{'id': cdt_bld + '.prefbase.toolchain.1',
				 'name': 'No ToolChain',
				 'resourceTypeBasedDiscovery': 'false',
				 'superClass': cdt_bld + '.prefbase.toolchain'})

		self.add(doc, toolChain, 'targetPlatform', {'binaryParser': 'org.eclipse.cdt.core.ELF', 'id': cdt_bld + '.prefbase.toolchain.1', 'name': ''})

		waf_build = '"%s" %s'%(waf_executable, eclipse.fun)
		waf_clean = '"%s" clean'%(waf_executable)
		self.add(doc, toolChain, 'builder',
					{'autoBuildTarget': waf_build,
					 'command': executable,
					 'enableAutoBuild': 'false',
					 'cleanBuildTarget': waf_clean,
					 'enableIncrementalBuild': 'true',
					 'id': cdt_bld + '.settings.default.builder.1',
					 'incrementalBuildTarget': waf_build,
					 'managedBuildOn': 'false',
					 'name': 'Gnu Make Builder',
					 'superClass': cdt_bld + '.settings.default.builder'})

		tool_index = 1;
		for tool_name in ("Assembly", "GNU C++", "GNU C"):
			tool = self.add(doc, toolChain, 'tool',
					{'id': cdt_bld + '.settings.holder.' + str(tool_index),
					 'name': tool_name,
					 'superClass': cdt_bld + '.settings.holder'})
			if cpppath or workspace_includes:
				incpaths = cdt_bld + '.settings.holder.incpaths'
				option = self.add(doc, tool, 'option',
						{'id': incpaths + '.' +  str(tool_index),
						 'name': 'Include Paths',
						 'superClass': incpaths,
						 'valueType': 'includePath'})
				for i in workspace_includes:
					self.add(doc, option, 'listOptionValue',
								{'builtIn': 'false',
								'value': '"${workspace_loc:/%s/%s}"'%(appname, i)})
				for i in cpppath:
					self.add(doc, option, 'listOptionValue',
								{'builtIn': 'false',
								'value': '"%s"'%(i)})
			if tool_name == "GNU C++" or tool_name == "GNU C":
				self.add(doc,tool,'inputType',{ 'id':'org.eclipse.cdt.build.core.settings.holder.inType.' + str(tool_index), \
					'languageId':'org.eclipse.cdt.core.gcc' if tool_name == "GNU C" else 'org.eclipse.cdt.core.g++','languageName':tool_name, \
					'sourceContentType':'org.eclipse.cdt.core.cSource,org.eclipse.cdt.core.cHeader', \
					'superClass':'org.eclipse.cdt.build.core.settings.holder.inType' })
			tool_index += 1

		if source_dirs:
			sourceEntries = self.add(doc, config, 'sourceEntries')
			for i in source_dirs:
				 self.add(doc, sourceEntries, 'entry',
							{'excluding': i,
							'flags': 'VALUE_WORKSPACE_PATH|RESOLVED',
							'kind': 'sourcePath',
							'name': ''})
				 self.add(doc, sourceEntries, 'entry',
							{
							'flags': 'VALUE_WORKSPACE_PATH|RESOLVED',
							'kind': 'sourcePath',
							'name': i})

		storageModule = self.add(doc, cconf, 'storageModule',
							{'moduleId': cdt_mk + '.buildtargets'})
		buildTargets = self.add(doc, storageModule, 'buildTargets')
		def addTargetWrap(name, runAll):
			return self.addTarget(doc, buildTargets, executable, name,
								'"%s" %s'%(waf_executable, name), runAll)
		addTargetWrap('configure', True)
		addTargetWrap('dist', False)
		addTargetWrap('install', False)
		for addTgt in self.env.ECLIPSE_EXTRA_TARGETS or []:
			addTargetWrap(addTgt, False)

		storageModule = self.add(doc, cproject, 'storageModule',
							{'moduleId': 'cdtBuildSystem',
							 'version': '4.0.0'})

		self.add(doc, storageModule, 'project', {'id': '%s.null.1'%appname, 'name': appname})

		storageModule = self.add(doc, cproject, 'storageModule',
							{'moduleId': 'org.eclipse.cdt.core.LanguageSettingsProviders'})

		storageModule = self.add(doc, cproject, 'storageModule',
							{'moduleId': 'scannerConfiguration'})

		doc.appendChild(cproject)
		return doc

	def impl_create_pydevproject(self, system_path, user_path):
		# create a pydevproject file
		doc = Document()
		doc.appendChild(doc.createProcessingInstruction('eclipse-pydev', 'version="1.0"'))
		pydevproject = doc.createElement('pydev_project')
		prop = self.add(doc, pydevproject,
					   'pydev_property',
					   'python %d.%d'%(sys.version_info[0], sys.version_info[1]))
		prop.setAttribute('name', 'org.python.pydev.PYTHON_PROJECT_VERSION')
		prop = self.add(doc, pydevproject, 'pydev_property', 'Default')
		prop.setAttribute('name', 'org.python.pydev.PYTHON_PROJECT_INTERPRETER')
		# add waf's paths
		wafadmin = [p for p in system_path if p.find('wafadmin') != -1]
		if wafadmin:
			prop = self.add(doc, pydevproject, 'pydev_pathproperty',
					{'name':'org.python.pydev.PROJECT_EXTERNAL_SOURCE_PATH'})
			for i in wafadmin:
				self.add(doc, prop, 'path', i)
		if user_path:
			prop = self.add(doc, pydevproject, 'pydev_pathproperty',
					{'name':'org.python.pydev.PROJECT_SOURCE_PATH'})
			for i in user_path:
				self.add(doc, prop, 'path', '/${PROJECT_DIR_NAME}/'+i)

		doc.appendChild(pydevproject)
		return doc

	def impl_create_javaproject(self, javasrcpath, javalibpath):
		# create a .classpath file for java usage
		doc = Document()
		javaproject = doc.createElement('classpath')
		if javasrcpath:
			for i in javasrcpath:
				self.add(doc, javaproject, 'classpathentry',
					{'kind': 'src', 'path': i})

		if javalibpath:
			for i in javalibpath:
				self.add(doc, javaproject, 'classpathentry',
					{'kind': 'lib', 'path': i})

		self.add(doc, javaproject, 'classpathentry', {'kind': 'con', 'path': 'org.eclipse.jdt.launching.JRE_CONTAINER'})
		self.add(doc, javaproject, 'classpathentry', {'kind': 'output', 'path': self.bldnode.name })
		doc.appendChild(javaproject)
		return doc

	def addDictionary(self, doc, parent, k, v):
		dictionary = self.add(doc, parent, 'dictionary')
		self.add(doc, dictionary, 'key', k)
		self.add(doc, dictionary, 'value', v)
		return dictionary

	def addTarget(self, doc, buildTargets, executable, name, buildTarget, runAllBuilders=True):
		target = self.add(doc, buildTargets, 'target',
						{'name': name,
						 'path': '',
						 'targetID': oe_cdt + '.build.MakeTargetBuilder'})
		self.add(doc, target, 'buildCommand', executable)
		self.add(doc, target, 'buildArguments', None)
		self.add(doc, target, 'buildTarget', buildTarget)
		self.add(doc, target, 'stopOnError', 'true')
		self.add(doc, target, 'useDefaultCommand', 'false')
		self.add(doc, target, 'runAllBuilders', str(runAllBuilders).lower())

	def add(self, doc, parent, tag, value = None):
		el = doc.createElement(tag)
		if (value):
			if type(value) == type(str()):
				el.appendChild(doc.createTextNode(value))
			elif type(value) == type(dict()):
				self.setAttributes(el, value)
		parent.appendChild(el)
		return el

	def setAttributes(self, node, attrs):
		for k, v in attrs.items():
			node.setAttribute(k, v)




#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2011 (ita)

"""
Latex processing using "biber"
"""

import os
from waflib import Task, Logs

from waflib.Tools import tex as texmodule

class tex(texmodule.tex):
	biber_fun, _ = Task.compile_fun('${BIBER} ${BIBERFLAGS} ${SRCFILE}',shell=False)
	biber_fun.__doc__ = """
	Execute the program **biber**
	"""

	def bibfile(self):
		return None

	def bibunits(self):
		self.env.env = {}
		self.env.env.update(os.environ)
		self.env.env.update({'BIBINPUTS': self.texinputs(), 'BSTINPUTS': self.texinputs()})
		self.env.SRCFILE = self.aux_nodes[0].name[:-4]

		if not self.env['PROMPT_LATEX']:
			self.env.append_unique('BIBERFLAGS', '--quiet')

		path = self.aux_nodes[0].abspath()[:-4] + '.bcf'
		if os.path.isfile(path):
			Logs.warn('calling biber')
			self.check_status('error when calling biber, check %s.blg for errors' % (self.env.SRCFILE), self.biber_fun())
		else:
			super(tex, self).bibfile()
			super(tex, self).bibunits()

class latex(tex):
	texfun, vars = Task.compile_fun('${LATEX} ${LATEXFLAGS} ${SRCFILE}', shell=False)
class pdflatex(tex):
	texfun, vars =  Task.compile_fun('${PDFLATEX} ${PDFLATEXFLAGS} ${SRCFILE}', shell=False)
class xelatex(tex):
	texfun, vars = Task.compile_fun('${XELATEX} ${XELATEXFLAGS} ${SRCFILE}', shell=False)

def configure(self):
	"""
	Almost the same as in tex.py, but try to detect 'biber'
	"""
	v = self.env
	for p in ' biber tex latex pdflatex xelatex bibtex dvips dvipdf ps2pdf makeindex pdf2ps'.split():
		try:
			self.find_program(p, var=p.upper())
		except self.errors.ConfigurationError:
			pass
	v['DVIPSFLAGS'] = '-Ppdf'




#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2005-2018 (ita)

"""
Runner.py: Task scheduling and execution
"""

import heapq, traceback
try:
	from queue import Queue, PriorityQueue
except ImportError:
	from Queue import Queue
	try:
		from Queue import PriorityQueue
	except ImportError:
		class PriorityQueue(Queue):
			def _init(self, maxsize):
				self.maxsize = maxsize
				self.queue = []
			def _put(self, item):
				heapq.heappush(self.queue, item)
			def _get(self):
				return heapq.heappop(self.queue)

from waflib import Utils, Task, Errors, Logs

GAP = 5
"""
Wait for at least ``GAP * njobs`` before trying to enqueue more tasks to run
"""

class PriorityTasks(object):
	def __init__(self):
		self.lst = []
	def __len__(self):
		return len(self.lst)
	def __iter__(self):
		return iter(self.lst)
	def __str__(self):
		return 'PriorityTasks: [%s]' % '\n  '.join(str(x) for x in self.lst)
	def clear(self):
		self.lst = []
	def append(self, task):
		heapq.heappush(self.lst, task)
	def pop(self):
		return heapq.heappop(self.lst)
	def extend(self, lst):
		if self.lst:
			for x in lst:
				self.append(x)
		else:
			if isinstance(lst, list):
				self.lst = lst
				heapq.heapify(lst)
			else:
				self.lst = lst.lst

class Consumer(Utils.threading.Thread):
	"""
	Daemon thread object that executes a task. It shares a semaphore with
	the coordinator :py:class:`waflib.Runner.Spawner`. There is one
	instance per task to consume.
	"""
	def __init__(self, spawner, task):
		Utils.threading.Thread.__init__(self)
		self.task = task
		"""Task to execute"""
		self.spawner = spawner
		"""Coordinator object"""
		self.daemon = True
		self.start()
	def run(self):
		"""
		Processes a single task
		"""
		try:
			if not self.spawner.master.stop:
				self.spawner.master.process_task(self.task)
		finally:
			self.spawner.sem.release()
			self.spawner.master.out.put(self.task)
			self.task = None
			self.spawner = None

class Spawner(Utils.threading.Thread):
	"""
	Daemon thread that consumes tasks from :py:class:`waflib.Runner.Parallel` producer and
	spawns a consuming thread :py:class:`waflib.Runner.Consumer` for each
	:py:class:`waflib.Task.Task` instance.
	"""
	def __init__(self, master):
		Utils.threading.Thread.__init__(self)
		self.master = master
		""":py:class:`waflib.Runner.Parallel` producer instance"""
		self.sem = Utils.threading.Semaphore(master.numjobs)
		"""Bounded semaphore that prevents spawning more than *n* concurrent consumers"""
		self.daemon = True
		self.start()
	def run(self):
		"""
		Spawns new consumers to execute tasks by delegating to :py:meth:`waflib.Runner.Spawner.loop`
		"""
		try:
			self.loop()
		except Exception:
			# Python 2 prints unnecessary messages when shutting down
			# we also want to stop the thread properly
			pass
	def loop(self):
		"""
		Consumes task objects from the producer; ends when the producer has no more
		task to provide.
		"""
		master = self.master
		while 1:
			task = master.ready.get()
			self.sem.acquire()
			if not master.stop:
				task.log_display(task.generator.bld)
			Consumer(self, task)

class Parallel(object):
	"""
	Schedule the tasks obtained from the build context for execution.
	"""
	def __init__(self, bld, j=2):
		"""
		The initialization requires a build context reference
		for computing the total number of jobs.
		"""

		self.numjobs = j
		"""
		Amount of parallel consumers to use
		"""

		self.bld = bld
		"""
		Instance of :py:class:`waflib.Build.BuildContext`
		"""

		self.outstanding = PriorityTasks()
		"""Heap of :py:class:`waflib.Task.Task` that may be ready to be executed"""

		self.postponed = PriorityTasks()
		"""Heap of :py:class:`waflib.Task.Task` which are not ready to run for non-DAG reasons"""

		self.incomplete = set()
		"""List of :py:class:`waflib.Task.Task` waiting for dependent tasks to complete (DAG)"""

		self.ready = PriorityQueue(0)
		"""List of :py:class:`waflib.Task.Task` ready to be executed by consumers"""

		self.out = Queue(0)
		"""List of :py:class:`waflib.Task.Task` returned by the task consumers"""

		self.count = 0
		"""Amount of tasks that may be processed by :py:class:`waflib.Runner.TaskConsumer`"""

		self.processed = 0
		"""Amount of tasks processed"""

		self.stop = False
		"""Error flag to stop the build"""

		self.error = []
		"""Tasks that could not be executed"""

		self.biter = None
		"""Task iterator which must give groups of parallelizable tasks when calling ``next()``"""

		self.dirty = False
		"""
		Flag that indicates that the build cache must be saved when a task was executed
		(calls :py:meth:`waflib.Build.BuildContext.store`)"""

		self.revdeps = Utils.defaultdict(set)
		"""
		The reverse dependency graph of dependencies obtained from Task.run_after
		"""

		self.spawner = None
		"""
		Coordinating daemon thread that spawns thread consumers
		"""
		if self.numjobs > 1:
			self.spawner = Spawner(self)

	def get_next_task(self):
		"""
		Obtains the next Task instance to run

		:rtype: :py:class:`waflib.Task.Task`
		"""
		if not self.outstanding:
			return None
		return self.outstanding.pop()

	def postpone(self, tsk):
		"""
		Adds the task to the list :py:attr:`waflib.Runner.Parallel.postponed`.
		The order is scrambled so as to consume as many tasks in parallel as possible.

		:param tsk: task instance
		:type tsk: :py:class:`waflib.Task.Task`
		"""
		self.postponed.append(tsk)

	def refill_task_list(self):
		"""
		Pulls a next group of tasks to execute in :py:attr:`waflib.Runner.Parallel.outstanding`.
		Ensures that all tasks in the current build group are complete before processing the next one.
		"""
		while self.count > self.numjobs * GAP:
			self.get_out()

		while not self.outstanding:
			if self.count:
				self.get_out()
				if self.outstanding:
					break
			elif self.postponed:
				try:
					cond = self.deadlock == self.processed
				except AttributeError:
					pass
				else:
					if cond:
						# The most common reason is conflicting build order declaration
						# for example: "X run_after Y" and "Y run_after X"
						# Another can be changing "run_after" dependencies while the build is running
						# for example: updating "tsk.run_after" in the "runnable_status" method
						lst = []
						for tsk in self.postponed:
							deps = [id(x) for x in tsk.run_after if not x.hasrun]
							lst.append('%s\t-> %r' % (repr(tsk), deps))
							if not deps:
								lst.append('\n  task %r dependencies are done, check its *runnable_status*?' % id(tsk))
						raise Errors.WafError('Deadlock detected: check the task build order%s' % ''.join(lst))
				self.deadlock = self.processed

			if self.postponed:
				self.outstanding.extend(self.postponed)
				self.postponed.clear()
			elif not self.count:
				if self.incomplete:
					for x in self.incomplete:
						for k in x.run_after:
							if not k.hasrun:
								break
						else:
							# dependency added after the build started without updating revdeps
							self.incomplete.remove(x)
							self.outstanding.append(x)
							break
					else:
						if self.stop or self.error:
							break
						raise Errors.WafError('Broken revdeps detected on %r' % self.incomplete)
				else:
					tasks = next(self.biter)
					ready, waiting = self.prio_and_split(tasks)
					self.outstanding.extend(ready)
					self.incomplete.update(waiting)
					self.total = self.bld.total()
					break

	def add_more_tasks(self, tsk):
		"""
		If a task provides :py:attr:`waflib.Task.Task.more_tasks`, then the tasks contained
		in that list are added to the current build and will be processed before the next build group.

		The priorities for dependent tasks are not re-calculated globally

		:param tsk: task instance
		:type tsk: :py:attr:`waflib.Task.Task`
		"""
		if getattr(tsk, 'more_tasks', None):
			more = set(tsk.more_tasks)
			groups_done = set()
			def iteri(a, b):
				for x in a:
					yield x
				for x in b:
					yield x

			# Update the dependency tree
			# this assumes that task.run_after values were updated
			for x in iteri(self.outstanding, self.incomplete):
				for k in x.run_after:
					if isinstance(k, Task.TaskGroup):
						if k not in groups_done:
							groups_done.add(k)
							for j in k.prev & more:
								self.revdeps[j].add(k)
					elif k in more:
						self.revdeps[k].add(x)

			ready, waiting = self.prio_and_split(tsk.more_tasks)
			self.outstanding.extend(ready)
			self.incomplete.update(waiting)
			self.total += len(tsk.more_tasks)

	def mark_finished(self, tsk):
		def try_unfreeze(x):
			# DAG ancestors are likely to be in the incomplete set
			# This assumes that the run_after contents have not changed
			# after the build starts, else a deadlock may occur
			if x in self.incomplete:
				# TODO remove dependencies to free some memory?
				# x.run_after.remove(tsk)
				for k in x.run_after:
					if not k.hasrun:
						break
				else:
					self.incomplete.remove(x)
					self.outstanding.append(x)

		if tsk in self.revdeps:
			for x in self.revdeps[tsk]:
				if isinstance(x, Task.TaskGroup):
					x.prev.remove(tsk)
					if not x.prev:
						for k in x.next:
							# TODO necessary optimization?
							k.run_after.remove(x)
							try_unfreeze(k)
						# TODO necessary optimization?
						x.next = []
				else:
					try_unfreeze(x)
			del self.revdeps[tsk]

		if hasattr(tsk, 'semaphore'):
			sem = tsk.semaphore
			try:
				sem.release(tsk)
			except KeyError:
				# TODO
				pass
			else:
				while sem.waiting and not sem.is_locked():
					# take a frozen task, make it ready to run
					x = sem.waiting.pop()
					self.add_task(x)

	def get_out(self):
		"""
		Waits for a Task that task consumers add to :py:attr:`waflib.Runner.Parallel.out` after execution.
		Adds more Tasks if necessary through :py:attr:`waflib.Runner.Parallel.add_more_tasks`.

		:rtype: :py:attr:`waflib.Task.Task`
		"""
		tsk = self.out.get()
		if not self.stop:
			self.add_more_tasks(tsk)
		self.mark_finished(tsk)

		self.count -= 1
		self.dirty = True
		return tsk

	def add_task(self, tsk):
		if hasattr(tsk, 'semaphore'):
			sem = tsk.semaphore
			try:
				sem.acquire(tsk)
			except IndexError:
				sem.waiting.add(tsk)
				return

		self.count += 1
		self.processed += 1
		if self.numjobs == 1:
			tsk.log_display(tsk.generator.bld)
			try:
				self.process_task(tsk)
			finally:
				self.out.put(tsk)
		else:
			self.ready.put(tsk)

	def process_task(self, tsk):
		"""
		Processes a task and attempts to stop the build in case of errors
		"""
		tsk.process()
		if tsk.hasrun != Task.SUCCESS:
			self.error_handler(tsk)

	def skip(self, tsk):
		"""
		Mark a task as skipped/up-to-date
		"""
		tsk.hasrun = Task.SKIPPED
		self.mark_finished(tsk)

	def cancel(self, tsk):
		"""
		Mark a task as failed because of unsatisfiable dependencies
		"""
		tsk.hasrun = Task.CANCELED
		self.mark_finished(tsk)

	def error_handler(self, tsk):
		"""
		Called when a task cannot be executed. The flag :py:attr:`waflib.Runner.Parallel.stop` is set,
		unless the build is executed with::

			$ waf build -k

		:param tsk: task instance
		:type tsk: :py:attr:`waflib.Task.Task`
		"""
		if not self.bld.keep:
			self.stop = True
		self.error.append(tsk)

	def task_status(self, tsk):
		"""
		Obtains the task status to decide whether to run it immediately or not.

		:return: the exit status, for example :py:attr:`waflib.Task.ASK_LATER`
		:rtype: integer
		"""
		try:
			return tsk.runnable_status()
		except Exception:
			self.processed += 1
			tsk.err_msg = traceback.format_exc()
			if not self.stop and self.bld.keep:
				self.skip(tsk)
				if self.bld.keep == 1:
					# if -k stop on the first exception, if -kk try to go as far as possible
					if Logs.verbose > 1 or not self.error:
						self.error.append(tsk)
					self.stop = True
				else:
					if Logs.verbose > 1:
						self.error.append(tsk)
				return Task.EXCEPTION

			tsk.hasrun = Task.EXCEPTION
			self.error_handler(tsk)

			return Task.EXCEPTION

	def start(self):
		"""
		Obtains Task instances from the BuildContext instance and adds the ones that need to be executed to
		:py:class:`waflib.Runner.Parallel.ready` so that the :py:class:`waflib.Runner.Spawner` consumer thread
		has them executed. Obtains the executed Tasks back from :py:class:`waflib.Runner.Parallel.out`
		and marks the build as failed by setting the ``stop`` flag.
		If only one job is used, then executes the tasks one by one, without consumers.
		"""
		self.total = self.bld.total()

		while not self.stop:

			self.refill_task_list()

			# consider the next task
			tsk = self.get_next_task()
			if not tsk:
				if self.count:
					# tasks may add new ones after they are run
					continue
				else:
					# no tasks to run, no tasks running, time to exit
					break

			if tsk.hasrun:
				# if the task is marked as "run", just skip it
				self.processed += 1
				continue

			if self.stop: # stop immediately after a failure is detected
				break

			st = self.task_status(tsk)
			if st == Task.RUN_ME:
				self.add_task(tsk)
			elif st == Task.ASK_LATER:
				self.postpone(tsk)
			elif st == Task.SKIP_ME:
				self.processed += 1
				self.skip(tsk)
				self.add_more_tasks(tsk)
			elif st == Task.CANCEL_ME:
				# A dependency problem has occurred, and the
				# build is most likely run with `waf -k`
				if Logs.verbose > 1:
					self.error.append(tsk)
				self.processed += 1
				self.cancel(tsk)

		# self.count represents the tasks that have been made available to the consumer threads
		# collect all the tasks after an error else the message may be incomplete
		while self.error and self.count:
			self.get_out()

		self.ready.put(None)
		if not self.stop:
			assert not self.count
			assert not self.postponed
			assert not self.incomplete

	def prio_and_split(self, tasks):
		"""
		Label input tasks with priority values, and return a pair containing
		the tasks that are ready to run and the tasks that are necessarily
		waiting for other tasks to complete.

		The priority system is really meant as an optional layer for optimization:
		dependency cycles are found quickly, and builds should be more efficient.
		A high priority number means that a task is processed first.

		This method can be overridden to disable the priority system::

			def prio_and_split(self, tasks):
				return tasks, []

		:return: A pair of task lists
		:rtype: tuple
		"""
		# to disable:
		#return tasks, []
		for x in tasks:
			x.visited = 0

		reverse = self.revdeps

		groups_done = set()
		for x in tasks:
			for k in x.run_after:
				if isinstance(k, Task.TaskGroup):
					if k not in groups_done:
						groups_done.add(k)
						for j in k.prev:
							reverse[j].add(k)
				else:
					reverse[k].add(x)

		# the priority number is not the tree depth
		def visit(n):
			if isinstance(n, Task.TaskGroup):
				return sum(visit(k) for k in n.next)

			if n.visited == 0:
				n.visited = 1

				if n in reverse:
					rev = reverse[n]
					n.prio_order = n.tree_weight + len(rev) + sum(visit(k) for k in rev)
				else:
					n.prio_order = n.tree_weight

				n.visited = 2
			elif n.visited == 1:
				raise Errors.WafError('Dependency cycle found!')
			return n.prio_order

		for x in tasks:
			if x.visited != 0:
				# must visit all to detect cycles
				continue
			try:
				visit(x)
			except Errors.WafError:
				self.debug_cycles(tasks, reverse)

		ready = []
		waiting = []
		for x in tasks:
			for k in x.run_after:
				if not k.hasrun:
					waiting.append(x)
					break
			else:
				ready.append(x)
		return (ready, waiting)

	def debug_cycles(self, tasks, reverse):
		tmp = {}
		for x in tasks:
			tmp[x] = 0

		def visit(n, acc):
			if isinstance(n, Task.TaskGroup):
				for k in n.next:
					visit(k, acc)
				return
			if tmp[n] == 0:
				tmp[n] = 1
				for k in reverse.get(n, []):
					visit(k, [n] + acc)
				tmp[n] = 2
			elif tmp[n] == 1:
				lst = []
				for tsk in acc:
					lst.append(repr(tsk))
					if tsk is n:
						# exclude prior nodes, we want the minimum cycle
						break
				raise Errors.WafError('Task dependency cycle in "run_after" constraints: %s' % ''.join(lst))
		for x in tasks:
			visit(x, [])




#! /usr/bin/env python
# encoding: UTF-8
# Thomas Nagy 2020 (ita)

from waflib import Utils
from waflib.Configure import conf

PTHREAD_CHECK = '''
#include <pthread.h>

static void* fun(void* params) {
    (void)params;
    return NULL;
}

int main(int argc, char **argv) {
    pthread_t thread;
    (void)argc;
    (void)argv;
    pthread_create(&thread, NULL, &fun, NULL);
    pthread_join(thread, NULL);
    return 0;
}
'''

@conf
def check_pthreads(self, mode=None):
	if not mode:
		mode = 'cxx' if self.env.CXX else 'c'

	if Utils.unversioned_sys_platform() == 'sunos':
		flags = ['-pthreads', '-lpthread', '-mt', '-pthread']
	else:
		flags = ['', '-lpthreads', '-Kthread', '-kthread', '-llthread', '-pthread', '-pthreads', '-mthreads', '-lpthread', '--thread-safe', '-mt']

	features = mode
	for flag in flags:
		self.env.stash()

		self.env[mode.upper() + 'FLAGS_PTHREAD'] = [flag]

		if flag:
			msg = ' -> Trying pthread compilation flag %s' % flag
			okmsg = 'needs %s' % flag
		else:
			msg = 'Checking if a pthread flag is necessary for compiling'
			okmsg = 'None'

		try:
			self.check(features=features, msg=msg, okmsg=okmsg, use='PTHREAD', fragment=PTHREAD_CHECK)
		except self.errors.ConfigurationError:
			self.env.revert()
			continue
		else:
			break
	else:
		self.fatal('Could not find a suitable pthreads flag for compiling')

	features = '%s %sprogram' % (mode, mode)
	for flag in flags:
		self.env.stash()

		self.env.LINKFLAGS_PTHREAD = [flag]

		if flag:
			msg = ' -> Trying pthread link flag %s' % flag
			okmsg = 'needs %s' % flag
		else:
			msg = 'Checking if a pthread flag is necessary for linking'
			okmsg = 'None'

		try:
			self.check(features=features, msg=msg, okmsg=okmsg, use='PTHREAD', fragment=PTHREAD_CHECK)
		except self.errors.ConfigurationError:
			self.env.revert()
			continue
		else:
			break
	else:
		self.fatal('Could not find a suitable pthreads flag for linking')


def configure(self):
	self.check_pthreads()




#! /usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2007-2010 (ita)

"""
Debugging helper for parallel compilation.

Copy it to your project and load it with::

	def options(opt):
		opt.load('parallel_debug', tooldir='.')
	def build(bld):
		...

The build will then output a file named pdebug.svg in the source directory.
"""

import re, sys, threading, time, traceback
try:
	from Queue import Queue
except:
	from queue import Queue
from waflib import Runner, Options, Task, Logs, Errors

SVG_TEMPLATE = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.0//EN" "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" version="1.0"
   x="${project.x}" y="${project.y}" width="${project.width}" height="${project.height}" id="svg602" xml:space="preserve">

<style type='text/css' media='screen'>
	g.over rect { stroke:#FF0000; fill-opacity:0.4 }
</style>

<script type='text/javascript'><![CDATA[
var svg  = document.getElementsByTagName('svg')[0];

svg.addEventListener('mouseover', function(e) {
	var g = e.target.parentNode;
	var x = document.getElementById('r_' + g.id);
	if (x) {
		g.setAttribute('class', g.getAttribute('class') + ' over');
		x.setAttribute('class', x.getAttribute('class') + ' over');
		showInfo(e, g.id, e.target.attributes.tooltip.value);
	}
}, false);

svg.addEventListener('mouseout', function(e) {
		var g = e.target.parentNode;
		var x = document.getElementById('r_' + g.id);
		if (x) {
			g.setAttribute('class', g.getAttribute('class').replace(' over', ''));
			x.setAttribute('class', x.getAttribute('class').replace(' over', ''));
			hideInfo(e);
		}
}, false);

function showInfo(evt, txt, details) {
${if project.tooltip}
	tooltip = document.getElementById('tooltip');

	var t = document.getElementById('tooltiptext');
	t.firstChild.data = txt + " " + details;

	var x = evt.clientX + 9;
	if (x > 250) { x -= t.getComputedTextLength() + 16; }
	var y = evt.clientY + 20;
	tooltip.setAttribute("transform", "translate(" + x + "," + y + ")");
	tooltip.setAttributeNS(null, "visibility", "visible");

	var r = document.getElementById('tooltiprect');
	r.setAttribute('width', t.getComputedTextLength() + 6);
${endif}
}

function hideInfo(evt) {
	var tooltip = document.getElementById('tooltip');
	tooltip.setAttributeNS(null,"visibility","hidden");
}
]]></script>

<!-- inkscape requires a big rectangle or it will not export the pictures properly -->
<rect
   x='${project.x}' y='${project.y}' width='${project.width}' height='${project.height}'
   style="font-size:10;fill:#ffffff;fill-opacity:0.01;fill-rule:evenodd;stroke:#ffffff;"></rect>

${if project.title}
  <text x="${project.title_x}" y="${project.title_y}"
    style="font-size:15px; text-anchor:middle; font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1px;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans">${project.title}</text>
${endif}


${for cls in project.groups}
  <g id='${cls.classname}'>
    ${for rect in cls.rects}
    <rect x='${rect.x}' y='${rect.y}' width='${rect.width}' height='${rect.height}' tooltip='${rect.name}' style="font-size:10;fill:${rect.color};fill-rule:evenodd;stroke:#000000;stroke-width:0.4;" />
    ${endfor}
  </g>
${endfor}

${for info in project.infos}
  <g id='r_${info.classname}'>
   <rect x='${info.x}' y='${info.y}' width='${info.width}' height='${info.height}' style="font-size:10;fill:${info.color};fill-rule:evenodd;stroke:#000000;stroke-width:0.4;" />
   <text x="${info.text_x}" y="${info.text_y}"
       style="font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1px;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans"
   >${info.text}</text>
  </g>
${endfor}

${if project.tooltip}
  <g transform="translate(0,0)" visibility="hidden" id="tooltip">
       <rect id="tooltiprect" y="-15" x="-3" width="1" height="20" style="stroke:black;fill:#edefc2;stroke-width:1"/>
       <text id="tooltiptext" style="font-family:Arial; font-size:12;fill:black;"> </text>
  </g>
${endif}

</svg>
"""

COMPILE_TEMPLATE = '''def f(project):
	lst = []
	def xml_escape(value):
		return value.replace("&", "&amp;").replace('"', "&quot;").replace("'", "&apos;").replace("<", "&lt;").replace(">", "&gt;")

	%s
	return ''.join(lst)
'''
reg_act = re.compile(r"(?P<backslash>\\)|(?P<dollar>\$\$)|(?P<subst>\$\{(?P<code>[^}]*?)\})", re.M)
def compile_template(line):

	extr = []
	def repl(match):
		g = match.group
		if g('dollar'):
			return "$"
		elif g('backslash'):
			return "\\"
		elif g('subst'):
			extr.append(g('code'))
			return "<<|@|>>"
		return None

	line2 = reg_act.sub(repl, line)
	params = line2.split('<<|@|>>')
	assert(extr)


	indent = 0
	buf = []
	app = buf.append

	def app(txt):
		buf.append(indent * '\t' + txt)

	for x in range(len(extr)):
		if params[x]:
			app("lst.append(%r)" % params[x])

		f = extr[x]
		if f.startswith(('if', 'for')):
			app(f + ':')
			indent += 1
		elif f.startswith('py:'):
			app(f[3:])
		elif f.startswith(('endif', 'endfor')):
			indent -= 1
		elif f.startswith(('else', 'elif')):
			indent -= 1
			app(f + ':')
			indent += 1
		elif f.startswith('xml:'):
			app('lst.append(xml_escape(%s))' % f[4:])
		else:
			#app('lst.append((%s) or "cannot find %s")' % (f, f))
			app('lst.append(str(%s))' % f)

	if extr:
		if params[-1]:
			app("lst.append(%r)" % params[-1])

	fun = COMPILE_TEMPLATE % "\n\t".join(buf)
	# uncomment the following to debug the template
	#for i, x in enumerate(fun.splitlines()):
	#	print i, x
	return Task.funex(fun)

# red   #ff4d4d
# green #4da74d
# lila  #a751ff

color2code = {
	'GREEN'  : '#4da74d',
	'YELLOW' : '#fefe44',
	'PINK'   : '#a751ff',
	'RED'    : '#cc1d1d',
	'BLUE'   : '#6687bb',
	'CYAN'   : '#34e2e2',
}

mp = {}
info = [] # list of (text,color)

def map_to_color(name):
	if name in mp:
		return mp[name]
	try:
		cls = Task.classes[name]
	except KeyError:
		return color2code['RED']
	if cls.color in mp:
		return mp[cls.color]
	if cls.color in color2code:
		return color2code[cls.color]
	return color2code['RED']

def process(self):
	m = self.generator.bld.producer
	try:
		# TODO another place for this?
		del self.generator.bld.task_sigs[self.uid()]
	except KeyError:
		pass

	self.generator.bld.producer.set_running(1, self)

	try:
		ret = self.run()
	except Exception:
		self.err_msg = traceback.format_exc()
		self.hasrun = Task.EXCEPTION

		# TODO cleanup
		m.error_handler(self)
		return

	if ret:
		self.err_code = ret
		self.hasrun = Task.CRASHED
	else:
		try:
			self.post_run()
		except Errors.WafError:
			pass
		except Exception:
			self.err_msg = traceback.format_exc()
			self.hasrun = Task.EXCEPTION
		else:
			self.hasrun = Task.SUCCESS
	if self.hasrun != Task.SUCCESS:
		m.error_handler(self)

	self.generator.bld.producer.set_running(-1, self)

Task.Task.process_back = Task.Task.process
Task.Task.process = process

old_start = Runner.Parallel.start
def do_start(self):
	try:
		Options.options.dband
	except AttributeError:
		self.bld.fatal('use def options(opt): opt.load("parallel_debug")!')

	self.taskinfo = Queue()
	old_start(self)
	if self.dirty:
		make_picture(self)
Runner.Parallel.start = do_start

lock_running = threading.Lock()
def set_running(self, by, tsk):
	with lock_running:
		try:
			cache = self.lock_cache
		except AttributeError:
			cache = self.lock_cache = {}

		i = 0
		if by > 0:
			vals = cache.values()
			for i in range(self.numjobs):
				if i not in vals:
					cache[tsk] = i
					break
		else:
			i = cache[tsk]
			del cache[tsk]

		self.taskinfo.put( (i, id(tsk), time.time(), tsk.__class__.__name__, self.processed, self.count, by, ",".join(map(str, tsk.outputs)))  )
Runner.Parallel.set_running = set_running

def name2class(name):
	return name.replace(' ', '_').replace('.', '_')

def make_picture(producer):
	# first, cast the parameters
	if not hasattr(producer.bld, 'path'):
		return

	tmp = []
	try:
		while True:
			tup = producer.taskinfo.get(False)
			tmp.append(list(tup))
	except:
		pass

	try:
		ini = float(tmp[0][2])
	except:
		return

	if not info:
		seen = []
		for x in tmp:
			name = x[3]
			if not name in seen:
				seen.append(name)
			else:
				continue

			info.append((name, map_to_color(name)))
		info.sort(key=lambda x: x[0])

	thread_count = 0
	acc = []
	for x in tmp:
		thread_count += x[6]
		acc.append("%d %d %f %r %d %d %d %s" % (x[0], x[1], x[2] - ini, x[3], x[4], x[5], thread_count, x[7]))

	data_node = producer.bld.path.make_node('pdebug.dat')
	data_node.write('\n'.join(acc))

	tmp = [lst[:2] + [float(lst[2]) - ini] + lst[3:] for lst in tmp]

	st = {}
	for l in tmp:
		if not l[0] in st:
			st[l[0]] = len(st.keys())
	tmp = [  [st[lst[0]]] + lst[1:] for lst in tmp ]
	THREAD_AMOUNT = len(st.keys())

	st = {}
	for l in tmp:
		if not l[1] in st:
			st[l[1]] = len(st.keys())
	tmp = [  [lst[0]] + [st[lst[1]]] + lst[2:] for lst in tmp ]


	BAND = Options.options.dband

	seen = {}
	acc = []
	for x in range(len(tmp)):
		line = tmp[x]
		id = line[1]

		if id in seen:
			continue
		seen[id] = True

		begin = line[2]
		thread_id = line[0]
		for y in range(x + 1, len(tmp)):
			line = tmp[y]
			if line[1] == id:
				end = line[2]
				#print id, thread_id, begin, end
				#acc.append(  ( 10*thread_id, 10*(thread_id+1), 10*begin, 10*end ) )
				acc.append( (BAND * begin, BAND*thread_id, BAND*end - BAND*begin, BAND, line[3], line[7]) )
				break

	if Options.options.dmaxtime < 0.1:
		gwidth = 1
		for x in tmp:
			m = BAND * x[2]
			if m > gwidth:
				gwidth = m
	else:
		gwidth = BAND * Options.options.dmaxtime

	ratio = float(Options.options.dwidth) / gwidth
	gwidth = Options.options.dwidth
	gheight = BAND * (THREAD_AMOUNT + len(info) + 1.5)


	# simple data model for our template
	class tobject(object):
		pass

	model = tobject()
	model.x = 0
	model.y = 0
	model.width = gwidth + 4
	model.height = gheight + 4

	model.tooltip = not Options.options.dnotooltip

	model.title = Options.options.dtitle
	model.title_x = gwidth / 2
	model.title_y = gheight + - 5

	groups = {}
	for (x, y, w, h, clsname, name) in acc:
		try:
			groups[clsname].append((x, y, w, h, name))
		except:
			groups[clsname] = [(x, y, w, h, name)]

	# groups of rectangles (else js highlighting is slow)
	model.groups = []
	for cls in groups:
		g = tobject()
		model.groups.append(g)
		g.classname = name2class(cls)
		g.rects = []
		for (x, y, w, h, name) in groups[cls]:
			r = tobject()
			g.rects.append(r)
			r.x = 2 + x * ratio
			r.y = 2 + y
			r.width = w * ratio
			r.height = h
			r.name = name
			r.color = map_to_color(cls)

	cnt = THREAD_AMOUNT

	# caption
	model.infos = []
	for (text, color) in info:
		inf = tobject()
		model.infos.append(inf)
		inf.classname = name2class(text)
		inf.x = 2 + BAND
		inf.y = 5 + (cnt + 0.5) * BAND
		inf.width = BAND/2
		inf.height = BAND/2
		inf.color = color

		inf.text = text
		inf.text_x = 2 + 2 * BAND
		inf.text_y = 5 + (cnt + 0.5) * BAND + 10

		cnt += 1

	# write the file...
	template1 = compile_template(SVG_TEMPLATE)
	txt = template1(model)

	node = producer.bld.path.make_node('pdebug.svg')
	node.write(txt)
	Logs.warn('Created the diagram %r', node)

def options(opt):
	opt.add_option('--dtitle', action='store', default='Parallel build representation for %r' % ' '.join(sys.argv),
		help='title for the svg diagram', dest='dtitle')
	opt.add_option('--dwidth', action='store', type=int, help='diagram width', default=800, dest='dwidth')
	opt.add_option('--dtime', action='store', type=float, help='recording interval in seconds', default=0.009, dest='dtime')
	opt.add_option('--dband', action='store', type=int, help='band width', default=22, dest='dband')
	opt.add_option('--dmaxtime', action='store', type=float, help='maximum time, for drawing fair comparisons', default=0, dest='dmaxtime')
	opt.add_option('--dnotooltip', action='store_true', help='disable tooltips', default=False, dest='dnotooltip')




#! /usr/bin/env python
# encoding: utf-8
# harald at klimachs.de

import re
from waflib import Utils
from waflib.Tools import fc,fc_config,fc_scan
from waflib.Configure import conf

from waflib.Tools.compiler_fc import fc_compiler
fc_compiler['linux'].append('fc_solstudio')

@conf
def find_solstudio(conf):
	"""Find the Solaris Studio compiler (will look in the environment variable 'FC')"""

	fc = conf.find_program(['sunf95', 'f95', 'sunf90', 'f90'], var='FC')
	conf.get_solstudio_version(fc)
	conf.env.FC_NAME = 'SOL'

@conf
def solstudio_flags(conf):
	v = conf.env
	v['FCFLAGS_fcshlib'] = ['-Kpic']
	v['FCFLAGS_DEBUG'] = ['-w3']
	v['LINKFLAGS_fcshlib'] = ['-G']
	v['FCSTLIB_MARKER'] = '-Bstatic'
	v['FCSHLIB_MARKER'] = '-Bdynamic'
	v['SONAME_ST']      = '-h %s'

@conf
def solstudio_modifier_platform(conf):
	dest_os = conf.env['DEST_OS'] or Utils.unversioned_sys_platform()
	solstudio_modifier_func = getattr(conf, 'solstudio_modifier_' + dest_os, None)
	if solstudio_modifier_func:
		solstudio_modifier_func()

@conf
def get_solstudio_version(conf, fc):
	"""Get the compiler version"""

	version_re = re.compile(r"Sun Fortran 95 *(?P<major>\d*)\.(?P<minor>\d*)", re.I).search
	cmd = fc + ['-V']

	out, err = fc_config.getoutput(conf,cmd,stdin=False)
	if out:
		match = version_re(out)
	else:
		match = version_re(err)
	if not match:
		conf.fatal('Could not determine the Sun Studio Fortran version.')
	k = match.groupdict()
	conf.env['FC_VERSION'] = (k['major'], k['minor'])

def configure(conf):
	conf.find_solstudio()
	conf.find_ar()
	conf.fc_flags()
	conf.fc_add_flags()
	conf.solstudio_flags()
	conf.solstudio_modifier_platform()




#! /usr/bin/env python
# encoding: utf-8
# harald at klimachs.de

"""
NEC SX Compiler for SX vector systems
"""

import re
from waflib import Errors, Utils
from waflib.Tools import ccroot,ar
from waflib.Configure import conf

from waflib.Tools import xlc # method xlc_common_flags
from waflib.Tools.compiler_c import c_compiler
c_compiler['linux'].append('c_nec')

@conf
def find_sxc(conf):
	cc = conf.find_program(['sxcc'], var='CC')
	conf.get_sxc_version(cc)
	conf.env.CC = cc
	conf.env.CC_NAME = 'sxcc'

@conf
def get_sxc_version(conf, fc):
	version_re = re.compile(r"C\+\+/SX\s*Version\s*(?P<major>\d*)\.(?P<minor>\d*)", re.I).search
	cmd = fc + ['-V']

	try:
		out, err = conf.cmd_and_log(cmd, output=0)
	except Errors.WafError:
		conf.fatal('Could not determine an sxcc version %r' % cmd)

	if out:
		match = version_re(out)
	else:
		match = version_re(err)
	if not match:
		conf.fatal('Could not determine the NEC C compiler version.')
	k = match.groupdict()
	conf.env['C_VERSION'] = (k['major'], k['minor'])

@conf
def sxc_common_flags(conf):
	v=conf.env
	v['CC_SRC_F']=[]
	v['CC_TGT_F']=['-c','-o']
	if not v['LINK_CC']:
		v['LINK_CC']=v['CC']
	v['CCLNK_SRC_F']=[]
	v['CCLNK_TGT_F']=['-o']
	v['CPPPATH_ST']='-I%s'
	v['DEFINES_ST']='-D%s'
	v['LIB_ST']='-l%s'
	v['LIBPATH_ST']='-L%s'
	v['STLIB_ST']='-l%s'
	v['STLIBPATH_ST']='-L%s'
	v['RPATH_ST']=''
	v['SONAME_ST']=[]
	v['SHLIB_MARKER']=[]
	v['STLIB_MARKER']=[]
	v['LINKFLAGS_cprogram']=['']
	v['cprogram_PATTERN']='%s'
	v['CFLAGS_cshlib']=['-fPIC']
	v['LINKFLAGS_cshlib']=['']
	v['cshlib_PATTERN']='lib%s.so'
	v['LINKFLAGS_cstlib']=[]
	v['cstlib_PATTERN']='lib%s.a'

def configure(conf):
	conf.find_sxc()
	conf.find_program('sxar',VAR='AR')
	conf.sxc_common_flags()
	conf.cc_load_tools()
	conf.cc_add_flags()
	conf.link_add_flags()



#! /usr/bin/env python
# encoding: utf-8
#
# written by Sylvain Rouquette, 2014

'''

This is an extra tool, not bundled with the default waf binary.
To add the cpplint tool to the waf file:
$ ./waf-light --tools=compat15,cpplint

this tool also requires cpplint for python.
If you have PIP, you can install it like this: pip install cpplint

When using this tool, the wscript will look like:

    def options(opt):
        opt.load('compiler_cxx cpplint')

    def configure(conf):
        conf.load('compiler_cxx cpplint')
        # optional, you can also specify them on the command line
        conf.env.CPPLINT_FILTERS = ','.join((
            '-whitespace/newline',      # c++11 lambda
            '-readability/braces',      # c++11 constructor
            '-whitespace/braces',       # c++11 constructor
            '-build/storage_class',     # c++11 for-range
            '-whitespace/blank_line',   # user pref
            '-whitespace/labels'        # user pref
            ))

    def build(bld):
        bld(features='cpplint', source='main.cpp', target='app')
        # add include files, because they aren't usually built
        bld(features='cpplint', source=bld.path.ant_glob('**/*.hpp'))
'''

import sys, re
import logging
from waflib import Errors, Task, TaskGen, Logs, Options, Node, Utils


critical_errors = 0
CPPLINT_FORMAT = '[CPPLINT] %(filename)s:\nline %(linenum)s, severity %(confidence)s, category: %(category)s\n%(message)s\n'
RE_EMACS = re.compile(r'(?P<filename>.*):(?P<linenum>\d+):  (?P<message>.*)  \[(?P<category>.*)\] \[(?P<confidence>\d+)\]')
CPPLINT_RE = {
    'waf': RE_EMACS,
    'emacs': RE_EMACS,
    'vs7': re.compile(r'(?P<filename>.*)\((?P<linenum>\d+)\):  (?P<message>.*)  \[(?P<category>.*)\] \[(?P<confidence>\d+)\]'),
    'eclipse': re.compile(r'(?P<filename>.*):(?P<linenum>\d+): warning: (?P<message>.*)  \[(?P<category>.*)\] \[(?P<confidence>\d+)\]'),
}
CPPLINT_STR = ('${CPPLINT} '
               '--verbose=${CPPLINT_LEVEL} '
               '--output=${CPPLINT_OUTPUT} '
               '--filter=${CPPLINT_FILTERS} '
               '--root=${CPPLINT_ROOT} '
               '--linelength=${CPPLINT_LINE_LENGTH} ')


def options(opt):
    opt.add_option('--cpplint-filters', type=str,
                   default='', dest='CPPLINT_FILTERS',
                   help='add filters to cpplint')
    opt.add_option('--cpplint-length', type=int,
                   default=80, dest='CPPLINT_LINE_LENGTH',
                   help='specify the line length (default: 80)')
    opt.add_option('--cpplint-level', default=1, type=int, dest='CPPLINT_LEVEL',
                   help='specify the log level (default: 1)')
    opt.add_option('--cpplint-break', default=5, type=int, dest='CPPLINT_BREAK',
                   help='break the build if error >= level (default: 5)')
    opt.add_option('--cpplint-root', type=str,
                   default='', dest='CPPLINT_ROOT',
                   help='root directory used to derive header guard')
    opt.add_option('--cpplint-skip', action='store_true',
                   default=False, dest='CPPLINT_SKIP',
                   help='skip cpplint during build')
    opt.add_option('--cpplint-output', type=str,
                   default='waf', dest='CPPLINT_OUTPUT',
                   help='select output format (waf, emacs, vs7, eclipse)')


def configure(conf):
    try:
        conf.find_program('cpplint', var='CPPLINT')
    except Errors.ConfigurationError:
        conf.env.CPPLINT_SKIP = True


class cpplint_formatter(Logs.formatter, object):
    def __init__(self, fmt):
        logging.Formatter.__init__(self, CPPLINT_FORMAT)
        self.fmt = fmt

    def format(self, rec):
        if self.fmt == 'waf':
            result = CPPLINT_RE[self.fmt].match(rec.msg).groupdict()
            rec.msg = CPPLINT_FORMAT % result
        if rec.levelno <= logging.INFO:
            rec.c1 = Logs.colors.CYAN
        return super(cpplint_formatter, self).format(rec)


class cpplint_handler(Logs.log_handler, object):
    def __init__(self, stream=sys.stderr, **kw):
        super(cpplint_handler, self).__init__(stream, **kw)
        self.stream = stream

    def emit(self, rec):
        rec.stream = self.stream
        self.emit_override(rec)
        self.flush()


class cpplint_wrapper(object):
    def __init__(self, logger, threshold, fmt):
        self.logger = logger
        self.threshold = threshold
        self.fmt = fmt

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if isinstance(exc_value, Utils.subprocess.CalledProcessError):
            messages = [m for m in exc_value.output.splitlines() 
                        if 'Done processing' not in m 
                        and 'Total errors found' not in m]
            for message in messages:
                self.write(message)
            return True

    def write(self, message):
        global critical_errors
        result = CPPLINT_RE[self.fmt].match(message)
        if not result:
            return
        level = int(result.groupdict()['confidence'])
        if level >= self.threshold:
            critical_errors += 1
        if level <= 2:
            self.logger.info(message)
        elif level <= 4:
            self.logger.warning(message)
        else:
            self.logger.error(message)


cpplint_logger = None
def get_cpplint_logger(fmt):
    global cpplint_logger
    if cpplint_logger:
        return cpplint_logger
    cpplint_logger = logging.getLogger('cpplint')
    hdlr = cpplint_handler()
    hdlr.setFormatter(cpplint_formatter(fmt))
    cpplint_logger.addHandler(hdlr)
    cpplint_logger.setLevel(logging.DEBUG)
    return cpplint_logger


class cpplint(Task.Task):
    color = 'PINK'

    def __init__(self, *k, **kw):
        super(cpplint, self).__init__(*k, **kw)

    def run(self):
        global critical_errors
        with cpplint_wrapper(get_cpplint_logger(self.env.CPPLINT_OUTPUT), self.env.CPPLINT_BREAK, self.env.CPPLINT_OUTPUT):
            params = {key: str(self.env[key]) for key in self.env if 'CPPLINT_' in key}
            if params['CPPLINT_OUTPUT'] == 'waf':
                params['CPPLINT_OUTPUT'] = 'emacs'
            params['CPPLINT'] = self.env.get_flat('CPPLINT')
            cmd = Utils.subst_vars(CPPLINT_STR, params)
            env = self.env.env or None
            Utils.subprocess.check_output(cmd + self.inputs[0].abspath(),
                                          stderr=Utils.subprocess.STDOUT,
                                          env=env, shell=True)
        return critical_errors

@TaskGen.extension('.h', '.hh', '.hpp', '.hxx')
def cpplint_includes(self, node):
    pass

@TaskGen.feature('cpplint')
@TaskGen.before_method('process_source')
def post_cpplint(self):
    if not self.env.CPPLINT_INITIALIZED:
        for key, value in Options.options.__dict__.items():
            if not key.startswith('CPPLINT_') or self.env[key]:
                continue
            self.env[key] = value
        self.env.CPPLINT_INITIALIZED = True

    if self.env.CPPLINT_SKIP:
        return

    if not self.env.CPPLINT_OUTPUT in CPPLINT_RE:
        return

    for src in self.to_list(getattr(self, 'source', [])):
        if isinstance(src, Node.Node):
            node = src
        else:
            node = self.path.find_or_declare(src)
        if not node:
            self.bld.fatal('Could not find %r' % src)
        self.create_task('cpplint', node)



#!/usr/bin/env python3
"""Tool used to discover Visual Studio side-by-side (SxS) installations.

Implements an alternative discovery mechanism for Visual Studio 2017 and
newer to support SxS toolsets and specific Windows SDK releases. This acts
as a replacement for msvc.py to allow developers to pin to specific Visual
Studio and Windows SDK releases.

Usage::

	$ waf configure --msvc_version='msvc 17.9' --winsdk-version='8.1'

or::

	def configure(ctx):
		ctx.env.MSVC_VERSIONS = ['msvc 17.9']
		ctx.env.WINSDK_VERSION = '8.1'
		ctx.load('msvc_toolset')
"""

import contextlib
import functools
import glob
import inspect
import json
import logging
import os
import re
import subprocess
import tempfile

try:  # noqa: SIM105
	from waflib import Configure, Context, Options
except ImportError:
	pass


@functools.total_ordering
class _VisualStudioToolset(object):  # noqa: UP004
	"""Class used to represent a Visual Studio toolset installation."""

	VS_INSTALL_ROOT = '{}\\Microsoft Visual Studio'.format(
		os.environ.get('PROGRAMFILES(X86)')
		or os.environ.get('PROGRAMFILES')
		or 'C:\\Program Files (x86)',
	)

	_discovered = None

	@classmethod
	def discover(cls, log_debug, runner=None, vswhere=None, include_prerelease=False):  # noqa: ANN001, ANN206, FBT002
		"""Memoize the Visual Studio toolsets installed.

		:param log_debug: Python logging.Logger.debug method
		:type log_debug: Callable[[str], None]
		:param runner: Method used to invoke a subprocess
		:type runner: Callable[[list(str)], str], optional
		:param vswhere: Path to vswhere executable
		:type vswhere: list(str), optional
		:param include_prerelease: Include prerelease versions
		:type include_prerelease: bool, optional
		:return: collection of toolsets installed
		:rtype: list[_VisualStudioToolset]
		"""
		if cls._discovered is None:
			run_process = runner or functools.partial(
				subprocess.check_output,
				universal_newlines=True,
			)
			args = (
				(vswhere or ['{}/Installer/vswhere.exe'.format(cls.VS_INSTALL_ROOT)])  # noqa: UP032
				+ [
					'-all',
					'-products',
					'*',
					'-format',
					'json',
					'-utf8',
					'-include',
					'packages',
				]
				+ (['-prerelease'] if include_prerelease else [])
			)
			listing = run_process(args)

			installations = json.loads(listing)

			vc_components = '{}(?:{})'.format(
				re.escape('Microsoft.VisualStudio.Component.VC.'),
				'|'.join(  # noqa: FLY002
					[
						r'Tools\.(?!14)',
						'140',
						'v141',
						r'(?:Tools\.)?14(?:\.\d+)+',
					],
				),
			)
			re_sxs_ids = re.compile(vc_components)

			vc_components = (
				'{}:{}'.format(install['installationVersion'], package['id'])
				for install in installations
				for package in install['packages']
				if re_sxs_ids.match(package['id'])
			)

			for component in vc_components:
				log_debug(component)

			sxs_minimum_version = 15
			installs = (
				(install['installationVersion'], install['installationPath'])
				for install in installations
				if next(int(i) for i in install['installationVersion'].split('.'))
				>= sxs_minimum_version
			)

			toolset_dirs = {
				cls(install_version, toolset_dir)
				for install_version, install_path in installs
				for toolset_dir in glob.iglob('{}/VC/Tools/MSVC/*'.format(install_path))  # noqa: PTH207, UP032
				if os.path.isdir('{}/bin'.format(toolset_dir))  # noqa: PTH112, UP032
			}

			cls._discovered = sorted(toolset_dirs)

		return cls._discovered

	def __init__(self, install_version, directory):  # noqa: ANN001, ANN204
		"""Construct an instance of the class.

		:param install_version: Visual Studio version containing the toolset
		:type install_version: str
		:param directory: Fully qualified path to the toolset directory
		:type directory: os.pathLike
		"""
		self.install_version = install_version
		self._path = directory
		self._vcvars_ver = os.path.basename(directory)  # noqa: PTH119
		self._vcvarsall = os.path.normpath(
			'{}/../../../Auxiliary/Build/vcvarsall.bat'.format(directory),  # noqa: UP032
		)

		version = tuple(int(i) for i in self._vcvars_ver.split('.'))
		# https://learn.microsoft.com/en-us/cpp/overview/compiler-versions?view=msvc-170  # noqa: ERA001,E501
		if version < (14, 20):
			self._visual_studio_version = (15, version[1] - 7)
		elif version < (14, 30):
			self._visual_studio_version = (
				16,
				version[1]
				- 20
				# 16.9 distinguished from 16.8 by build number
				# this also accounts for the minor version not
				# being updated for 16.10
				+ bool(version >= (14, 28, 29910))
				# 16.11 distinguished from 16.10 by build number
				+ bool(version >= (14, 29, 30129)),
			)
		else:
			self._visual_studio_version = (17, version[1] - 30)

	def __eq__(self, other):  # noqa: ANN001, ANN204
		"""Return if the instance is equal to another instance of this class.

		:param other: Another toolset to compare against
		:type other: _VisualStudioToolset
		:return: True if the toolsets represent the same Visual Studio version, False otherwise
		:rtype: bool
		"""
		return self._visual_studio_version == other._visual_studio_version

	def __lt__(self, other):  # noqa: ANN001, ANN204
		"""Return if the instance is less than another instance of this class.

		:param other: Another toolset to compare against
		:type other: _VisualStudioToolset
		:return: True if the toolsets represent an older Visual Studio version, False otherwise
		:rtype: bool
		"""
		return self._visual_studio_version < other._visual_studio_version

	def __hash__(self):  # noqa: ANN204
		"""Utilize _path member as class hash.

		:return: Python hash of the toolset directory
		:rtype: int
		"""
		return hash(self._path)

	def __str__(self):  # noqa: ANN204
		"""Obtain the string representation of the class.

		:return: A uniquely identified string representation for the toolset
		:rtype: str
		"""
		return '{:<16s} {:<6s} {}'.format(
			'{}:'.format(self.install_version),  # noqa: UP032
			'{}:'.format(self.visual_studio_version),  # noqa: UP032
			self.vc_version,
		)

	@property
	def visual_studio_version(self):  # noqa: ANN202
		"""Obtain the Visual Studio version associated with this toolset.

		:return: Visual Studio major and minor version.
		:rtype: str
		"""
		return '{}.{}'.format(*self._visual_studio_version)

	@property
	def vcvarsall(self):  # noqa: ANN202
		"""Obtain the configuration script.

		:return: Fully qualified path to vcvarsall.bat
		:rtype: str
		"""
		return self._vcvarsall

	@property
	def vc_version(self):  # noqa: ANN202
		"""Obtain the vc_version option.

		:return: vc_version used to configure toolset in vcvarsall.bat
		:rtype: str
		"""
		return self._vcvars_ver

	def cl(self):  # noqa: ANN202
		"""Run and return the compiler output.

		:return: CL output on success, or vcvarsall output on error
		:rtype: list(str)
		"""
		contents = inspect.cleandoc(
			"""
			@echo off
			call "{}" amd64 -vcvars_ver={}
			if not "%ERRORLEVEL%"=="0" exit /b %ERRORLEVEL%
			cl
			""",
		).format(self.vcvarsall, self._vcvars_ver)

		with tempfile.NamedTemporaryFile('w', suffix='.cmd', delete=False) as tmp:
			try:
				tmp.write(contents)
				tmp.close()
				process = subprocess.Popen(  # noqa: S603
					[tmp.name],
					stdout=subprocess.PIPE,
					stderr=subprocess.PIPE,
					universal_newlines=True,
					shell=False,
				)
				stdout, stderr = process.communicate()
				if process.returncode:
					return stdout.splitlines() + stderr.splitlines()
				return stderr.splitlines()
			finally:
				os.unlink(tmp.name)  # noqa: PTH108


def options(ctx):  # noqa: ANN001, ANN201
	"""Configure the Microsoft compiler(s) Waf options.

	:param ctx: Waf context associated with option handling
	:type ctx: waflib.Options.OptionContext
	"""
	ctx.load('msvc')

	winsdk_default = re.findall(
		r'(8\.1|10(?:\.\d+)+)',
		os.environ.get('WINDOWSSDKVERSION', ''),
	) or ['']

	ctx.add_option(
		'--winsdk-version',
		type=str,
		help='Windows SDK Versions, eg: 10.0.19041.0',
		default=winsdk_default[0],
	)


@contextlib.contextmanager
def _use_vs_toolsets_discovery():  # noqa: ANN202
	"""Patch the Waf msvc tools VS2017+ discovery."""
	msvc = Context.load_tool('msvc')

	def gather_vswhere_versions(ctx, versions):  # noqa: ANN001, ANN202
		"""Detect the Microsoft SxS compiler(s) for Waf.

		:param ctx: Waf context associated with tool configuration
		:type ctx: waflib.Configure.ConfigurationContext
		:param versions: The upstream msvc versions dictionary
		:type versions: dict
		"""
		msvc.gather_vswhere_versions(ctx, versions)
		log_debug = functools.partial(
			logging.getLogger('waflib').debug,
			extra={'zone': 'msvc_toolset'},
		)

		for toolset in _VisualStudioToolset.discover(
			log_debug,
			ctx.cmd_and_log,
			vswhere=ctx.env.VSWHERE,
			include_prerelease=ctx.options.msvc_prerelease,
		):
			waf_version = 'msvc {}'.format(toolset.visual_studio_version)  # noqa: UP032
			if waf_version in versions:
				log_debug("Replacing msvc's '%s' with msvc_toolset's '%s'", waf_version, toolset)
			versions[waf_version] = {
				target: msvc.target_compiler(
					ctx,
					'msvc',
					realtarget,
					toolset.visual_studio_version,
					'{} {} -vcvars_ver={}'.format(
						target,
						getattr(Options.options, 'winsdk_version', '')
						or ctx.env.WINSDK_VERSION
						or '',
						toolset.vc_version,
					),
					toolset.vcvarsall,
				)
				for target, realtarget in msvc.all_msvc_platforms[::-1]
			}

	try:
		Configure.conf(gather_vswhere_versions)
		yield
	finally:
		Configure.conf(msvc.gather_vswhere_versions)


def configure(ctx):  # noqa: ANN001, ANN201
	"""Configure the Microsoft compiler(s) in Waf.

	:param ctx: Waf context associated with tool configuration
	:type ctx: waflib.Configure.ConfigurationContext
	"""
	with _use_vs_toolsets_discovery():
		ctx.load('msvc')


if __name__ == '__main__':
	logger = logging.getLogger('msvc_toolset')
	logging.basicConfig(
		format='%(message)s',
		level=getattr(logging, os.environ.get('LOG_LEVEL', 'info').upper()),
	)

	for toolset in _VisualStudioToolset.discover(logger.debug):
		logger.info(toolset)
		cl_info = '\n    '.join(line for line in toolset.cl() if line)
		logger.info('    %s', cl_info)



#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2011 (ita)

"""
Experimental F# stuff

FSC="mono /path/to/fsc.exe" waf configure build
"""

from waflib import Utils, Task
from waflib.TaskGen import before_method, after_method, feature
from waflib.Tools import ccroot, cs

ccroot.USELIB_VARS['fsc'] = set(['CSFLAGS', 'ASSEMBLIES', 'RESOURCES'])

@feature('fs')
@before_method('process_source')
def apply_fsc(self):
	cs_nodes = []
	no_nodes = []
	for x in self.to_nodes(self.source):
		if x.name.endswith('.fs'):
			cs_nodes.append(x)
		else:
			no_nodes.append(x)
	self.source = no_nodes

	bintype = getattr(self, 'type', self.gen.endswith('.dll') and 'library' or 'exe')
	self.cs_task = tsk = self.create_task('fsc', cs_nodes, self.path.find_or_declare(self.gen))
	tsk.env.CSTYPE = '/target:%s' % bintype
	tsk.env.OUT    = '/out:%s' % tsk.outputs[0].abspath()

	inst_to = getattr(self, 'install_path', bintype=='exe' and '${BINDIR}' or '${LIBDIR}')
	if inst_to:
		# note: we are making a copy, so the files added to cs_task.outputs won't be installed automatically
		mod = getattr(self, 'chmod', bintype=='exe' and Utils.O755 or Utils.O644)
		self.install_task = self.add_install_files(install_to=inst_to, install_from=self.cs_task.outputs[:], chmod=mod)

feature('fs')(cs.use_cs)
after_method('apply_fsc')(cs.use_cs)

feature('fs')(cs.debug_cs)
after_method('apply_fsc', 'use_cs')(cs.debug_cs)

class fsc(Task.Task):
	"""
	Compile F# files
	"""
	color   = 'YELLOW'
	run_str = '${FSC} ${CSTYPE} ${CSFLAGS} ${ASS_ST:ASSEMBLIES} ${RES_ST:RESOURCES} ${OUT} ${SRC}'

def configure(conf):
	"""
	Find a F# compiler, set the variable FSC for the compiler and FS_NAME (mono or fsc)
	"""
	conf.find_program(['fsc.exe', 'fsharpc'], var='FSC')
	conf.env.ASS_ST = '/r:%s'
	conf.env.RES_ST = '/resource:%s'

	conf.env.FS_NAME = 'fsc'
	if str(conf.env.FSC).lower().find('fsharpc') > -1:
		conf.env.FS_NAME = 'mono'




#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2006-2018 (ita)
# Ralf Habacker, 2006 (rh)

from waflib import Errors
from waflib.Tools import ccroot, ar
from waflib.Configure import conf

@conf
def find_scc(conf):
	"""
	Detects the Sun C compiler
	"""
	v = conf.env
	cc = conf.find_program('cc', var='CC')
	try:
		conf.cmd_and_log(cc + ['-flags'])
	except Errors.WafError:
		conf.fatal('%r is not a Sun compiler' % cc)
	v.CC_NAME = 'sun'
	conf.get_suncc_version(cc)

@conf
def scc_common_flags(conf):
	"""
	Flags required for executing the sun C compiler
	"""
	v = conf.env

	v.CC_SRC_F            = []
	v.CC_TGT_F            = ['-c', '-o', '']

	if not v.LINK_CC:
		v.LINK_CC = v.CC

	v.CCLNK_SRC_F         = ''
	v.CCLNK_TGT_F         = ['-o', '']
	v.CPPPATH_ST          = '-I%s'
	v.DEFINES_ST          = '-D%s'

	v.LIB_ST              = '-l%s' # template for adding libs
	v.LIBPATH_ST          = '-L%s' # template for adding libpaths
	v.STLIB_ST            = '-l%s'
	v.STLIBPATH_ST        = '-L%s'

	v.SONAME_ST           = '-Wl,-h,%s'
	v.SHLIB_MARKER        = '-Bdynamic'
	v.STLIB_MARKER        = '-Bstatic'

	v.cprogram_PATTERN    = '%s'

	v.CFLAGS_cshlib       = ['-xcode=pic32', '-DPIC']
	v.LINKFLAGS_cshlib    = ['-G']
	v.cshlib_PATTERN      = 'lib%s.so'

	v.LINKFLAGS_cstlib    = ['-Bstatic']
	v.cstlib_PATTERN      = 'lib%s.a'

def configure(conf):
	conf.find_scc()
	conf.find_ar()
	conf.scc_common_flags()
	conf.cc_load_tools()
	conf.cc_add_flags()
	conf.link_add_flags()




#! /usr/bin/env python
# Thomas Nagy, 2011 (ita)

"""
Create _moc.cpp files

The builds are 30-40% faster when .moc files are directly included,
so the usage of this tool is discouraged.

def configure(conf):
	conf.load('compiler_cxx qt5')
	conf.load('slow_qt')

See playground/slow_qt/wscript for a complete example,
and run with "waf --zones=slow_qt" to display the moc files that should be generated
"""

from waflib.TaskGen import extension
from waflib import Task, Logs
import waflib.Tools.qt5
import waflib.Tools.cxx

@extension(*waflib.Tools.qt5.EXT_QT5)
def cxx_hook(self, node):
	return self.create_compiled_task('cxx_qt', node)

class cxx_qt(Task.classes['cxx']):
	def runnable_status(self):
		ret = Task.classes['cxx'].runnable_status(self)
		if ret != Task.ASK_LATER and not getattr(self, 'moc_done', None):
			try:
				cache = self.generator.moc_cache
			except AttributeError:
				cache = self.generator.moc_cache = {}

			deps = self.generator.bld.node_deps[self.uid()]
			for x in [self.inputs[0]] + deps:
				if x.read().find('Q_OBJECT') > 0:

					# process "foo.h -> foo.moc" only if "foo.cpp" is in the sources for the current task generator
					# this code will work because it is in the main thread (runnable_status)
					if x.name.rfind('.') > -1: # a .h file...
						name = x.name[:x.name.rfind('.')]
						for tsk in self.generator.compiled_tasks:
							if tsk.inputs and tsk.inputs[0].name.startswith(name):
								break
						else:
							# no corresponding file, continue
							continue

					# the file foo.cpp could be compiled for a static and a shared library
					# one workaround is to use a %number in the name
					#cxx_node = x.parent.get_bld().make_node(x.name.replace('.', '_') + '_%d_moc.cpp' % self.generator.idx)

					# another workaround is to add the target name
					cxx_node = x.parent.get_bld().make_node(x.name.replace('.', '_') + '_%s_moc.cpp' % self.generator.name)
					if cxx_node in cache:
						continue
					cache[cxx_node] = self

					Logs.debug('slow_qt: will create a file named %s', cxx_node.abspath())

					tsk = Task.classes['moc'](env=self.env, generator=self.generator)
					tsk.set_inputs(x)
					tsk.set_outputs(cxx_node)

					if x.name.endswith('.cpp'):
						gen = self.generator.bld.producer
						gen.outstanding.append(tsk)
						gen.total += 1
						self.set_run_after(tsk)
					else:
						cxxtsk = Task.classes['cxx'](env=self.env, generator=self.generator)
						cxxtsk.set_inputs(tsk.outputs)
						cxxtsk.set_outputs(cxx_node.change_ext('.o'))
						cxxtsk.set_run_after(tsk)

						try:
							self.more_tasks.extend([tsk, cxxtsk])
						except AttributeError:
							self.more_tasks = [tsk, cxxtsk]

						try:
							link = self.generator.link_task
						except AttributeError:
							pass
						else:
							link.set_run_after(cxxtsk)
							link.inputs.extend(cxxtsk.outputs)
							link.inputs.sort(key=lambda x: x.abspath())

			self.moc_done = True

		for t in self.run_after:
			if not t.hasrun:
				return Task.ASK_LATER

		return ret



#!/usr/bin/env python
# encoding: utf-8
# vim: sw=4 ts=4 noexpandtab

"""
LLVM Clang-CL support.

Clang-CL is supposed to be a drop-in replacement for MSVC CL, but also serves
well as a cross compiler for Windows from Linux (provided you have set up the
environment). Requires Visual Studio 2015+ to be installed.

On Windows it uses (most) MSVC tools.

Usage:
	$ waf configure
Or:
	$ LLVM_PATH=C:\\Program Files\\LLVM\\bin waf configure
Or:
	def configure(self):
		self.env.LLVM_PATH = 'C:\\Program Files\\LLVM\\bin'
		self.load('clang_cl')
"""

import os

from waflib import Utils, Errors, Logs
from waflib.Configure import conf
from waflib.Tools import msvc

def options(opt):
	msvc.options(opt)

@conf
def get_llvm_paths(self):
	llvm_path = []
	if Utils.is_win32:
		try:
			llvm_key = Utils.winreg.OpenKey(Utils.winreg.HKEY_LOCAL_MACHINE,'SOFTWARE\\Wow6432Node\\LLVM\\LLVM')
		except OSError:
			try:
				llvm_key = Utils.winreg.OpenKey(Utils.winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\LLVM\\LLVM')
			except OSError:
				llvm_key = None

		if llvm_key:
			llvm_dir, _ = Utils.winreg.QueryValueEx(llvm_key, '')
			if llvm_dir:
				llvm_path.append(os.path.join(llvm_dir, 'bin'))

	tmp = self.environ.get('LLVM_PATH') or self.env.LLVM_PATH
	if tmp:
		llvm_path.append(tmp)
	llvm_path += self.env.PATH
	return llvm_path

@conf
def find_clang_cl(self):
	"""
	Find the program clang-cl.
	"""
	del(self.env.CC)
	del(self.env.CXX)

	paths = self.get_llvm_paths()
	cc = self.find_program('clang-cl', var='CC', path_list=paths)
	self.env.CC = self.env.CXX = cc
	self.env.CC_NAME_SECONDARY = self.env.CXX_NAME_SECONDARY = 'clang'

	if not Utils.is_win32:
		self.env.MSVC_COMPILER = 'msvc'
		self.env.MSVC_VERSION = 19

		if not self.env.LINK_CXX:
			self.find_program('lld-link', path_list=paths, errmsg='lld-link was not found (linker)', var='LINK_CXX')

		if not self.env.LINK_CC:
			self.env.LINK_CC = self.env.LINK_CXX

@conf
def find_llvm_tools(self):
	"""
	Find the librarian, manifest tool, and resource compiler.
	"""
	self.env.CC_NAME = self.env.CXX_NAME = 'msvc'

	paths = self.get_llvm_paths()
	llvm_path = self.environ.get('LLVM_PATH') or self.env.LLVM_PATH
	if llvm_path:
		paths = [llvm_path] + self.env.PATH
	else:
		paths = self.env.PATH

	if not self.env.AR:
		stliblink = self.find_program('llvm-lib', path_list=paths, var='AR')
		if not stliblink:
			self.fatal('Unable to find required program "llvm-lib"')

		self.env.ARFLAGS = ['/nologo']

	# We assume clang_cl to only be used with relatively new MSVC installations.
	self.env.MSVC_MANIFEST = True
	self.find_program('llvm-mt', path_list=paths, var='MT')
	self.env.MTFLAGS = ['/nologo']

	try:
		self.load('winres')
	except Errors.ConfigurationError:
		Logs.warn('Resource compiler not found. Compiling resource file is disabled')

def configure(self):
	if Utils.is_win32:
		self.autodetect(True)
		self.find_msvc()
	else:
		self.find_llvm_tools()

	self.find_clang_cl()
	self.msvc_common_flags()
	self.cc_load_tools()
	self.cxx_load_tools()
	self.cc_add_flags()
	self.cxx_add_flags()
	self.link_add_flags()
	self.visual_studio_add_flags()
